# SPDX-License-Identifier: BSD-2-Clause

from bisect        import bisect_left, bisect_right
from typing        import Generator, Iterable, Optional, Union

from ...util.units import bits_for

__all__ = (
	'MemoryMap',
	'ResourceInfo',
)

class _RangeMap:
	'''
	Range map.

	A range map is a mapping from non-overlapping ranges to arbitrary values.

	'''

	def __init__(self) -> None:
		self._keys   = []
		self._values = dict()
		self._starts = []
		self._stops  = []

	def insert(self, key: range, value: object):
		assert isinstance(key, range)
		assert not self.overlaps(key)

		start_idx = bisect_right(self._starts, key.start)
		stop_idx  = bisect_left(self._stops, key.stop)
		assert start_idx == stop_idx

		self._starts.insert(start_idx, key.start)
		self._stops.insert(stop_idx, key.stop)
		self._keys.insert(start_idx, key)
		self._values[key] = value

	def get(self, point: int) -> object:
		point_idx = bisect_right(self._stops, point)
		if point_idx < len(self._keys):
			point_range = self._keys[point_idx]
			if point >= point_range.start and point < point_range.stop:
				return self._values[point_range]

	def overlaps(self, key: range) -> list[object]:
		start_idx = bisect_right(self._stops, key.start)
		stop_idx  = bisect_left(self._starts, key.stop)
		return [self._values[key] for key in self._keys[start_idx:stop_idx]]

	def items(self) -> Generator[tuple[range, object], None, None]:
		for key in self._keys:
			yield (key, self._values[key])


class ResourceInfo:
	'''
	Resource metadata.

	A wrapper class for resource objects, with their assigned name and address range.

	Parameters
	----------
	resource : object
		Arbitrary object representing a resource. See :meth:`MemoryMap.add_resource` for details.
	name : iter(str)
		Name assigned to the resource. It is prefixed by the name of each window sitting between
		the resource and the memory map from which this :class:`ResourceInfo` was obtained.
		See :meth:`MemoryMap.add_window` for details.
	start : int
		Start of the address range assigned to the resource.
	end : int
		End of the address range assigned to the resource.
	width : int
		Amount of data bits accessed at each address. It may be equal to the data width of the
		memory map from which this :class:`ResourceInfo` was obtained, or less if the resource
		is located behind a window that uses sparse addressing.

	'''

	def __init__(
		self, resource: object, name: Union[str, Iterable[str]], start: int, end: int, width: int
	) -> None:
		if isinstance(name, str):
			name = (name,)
		if not name or not all(isinstance(part, str) and part for part in name):
			raise TypeError(f'Name must be a non-empty sequence of non-empty strings, not {name!r}')
		if not isinstance(start, int) or start < 0:
			raise TypeError(f'Start address must be a non-negative integer, not {start!r}')
		if not isinstance(end, int) or end <= start:
			raise TypeError(f'End address must be an integer greater than the start address, not {end!r}')
		if not isinstance(width, int) or width < 0:
			raise TypeError(f'Width must be a non-negative integer, not {width!r}')

		self._resource = resource
		self._name     = tuple(name)
		self._start    = start
		self._end      = end
		self._width    = width

	@property
	def resource(self) -> object:
		return self._resource

	@property
	def name(self) -> tuple[str]:
		return self._name

	@property
	def start(self) -> int:
		return self._start

	@property
	def end(self) -> int:
		return self._end

	@property
	def width(self) -> int:
		return self._width


class MemoryMap:
	'''
	Memory map.

	A memory map is a hierarchical description of an address space, describing the structure of
	address decoders of peripherals as well as bus bridges. It is built by adding resources
	(range allocations for registers, memory, etc) and windows (range allocations for bus bridges),
	and can be queried later to determine the address of any given resource from a specific
	vantage point in the design.

	Address assignment
	------------------

	To simplify address assignment, each memory map has an implicit next address, starting at 0.
	If a resource or a window is added without specifying an address explicitly, the implicit next
	address is used. In any case, the implicit next address is set to the address immediately
	following the newly added resource or window.

	Parameters
	----------
	addr_width : int
		Address width.
	data_width : int
		Data width.
	alignment : log2 of int
		Range alignment. Each added resource and window will be placed at an address that is
		a multiple of ``2 ** alignment``, and its size will be rounded up to be a multiple of
		``2 ** alignment``.
	name : str
		Name of the address range. Optional.

	'''

	def __init__(
		self, *, addr_width: int, data_width: int, alignment: int = 0, name: Optional[str] = None
	) -> None:
		if not isinstance(addr_width, int) or addr_width <= 0:
			raise ValueError(f'Address width must be a positive integer, not {addr_width!r}')
		if not isinstance(data_width, int) or data_width <= 0:
			raise ValueError(f'Data width must be a positive integer, not {data_width!r}')
		if not isinstance(alignment, int) or alignment < 0:
			raise ValueError(f'Alignment must be a non-negative integer, not {alignment!r}')
		if name is not None and not (isinstance(name, str) and name):
			raise ValueError(f'Name must be a non-empty string, not {name!r}')

		self._addr_width = addr_width
		self._data_width = data_width
		self._alignment  = alignment
		self._name       = name

		self._ranges     = _RangeMap()
		self._resources  = dict()
		self._windows    = dict()
		self._namespace  = dict()

		self._next_addr  = 0
		self._frozen     = False

	@property
	def addr_width(self) -> int:
		return self._addr_width

	@addr_width.setter
	def addr_width(self, addr_width: int) -> None:
		if self._frozen:
			raise ValueError('Memory map has been frozen. Address width cannot be extended further')
		if not isinstance(addr_width, int) or addr_width <= 0:
			raise ValueError(f'Address width must be a positive integer, not {addr_width!r}')
		if addr_width < self._addr_width:
			raise ValueError(
				f'Address width {addr_width!r} must not be less than its previous value {self._addr_width!r}, '
				'because resources that were previously added may not fit anymore'
			)
		self._addr_width = addr_width

	@property
	def data_width(self) -> int:
		return self._data_width

	@property
	def alignment(self) -> int:
		return self._alignment

	@property
	def name(self) -> Optional[str]:
		return self._name

	def freeze(self) -> None:
		'''
		Freeze the memory map.

		Once the memory map is frozen, its visible state becomes immutable. Resources and windows
		cannot be added anymore, and its address width cannot be extended further.

		'''

		self._frozen = True

	@staticmethod
	def _align_up(value: int, alignment: int) -> int:
		if value % (1 << alignment) != 0:
			value += (1 << alignment) - (value % (1 << alignment))
		return value

	def align_to(self, alignment: int) -> int:
		'''
		Align the implicit next address.

		Arguments
		---------
		alignment : log2 of int
			Address alignment. The start of the implicit next address will be a multiple of
			``2 ** max(alignment, self.alignment)``.

		Return value
		------------
		Implicit next address.

		'''

		if not isinstance(alignment, int) or alignment < 0:
			raise ValueError(f'Alignment must be a non-negative integer, not {alignment!r}')
		self._next_addr = self._align_up(self._next_addr, max(alignment, self.alignment))
		return self._next_addr

	def _compute_addr_range(
		self, addr: int, size: int, step: int = 1, *, alignment: int, extend: bool
	) -> range:
		if addr is not None:
			if not isinstance(addr, int) or addr < 0:
				raise ValueError(f'Address must be a non-negative integer, not {addr!r}')
			if addr % (1 << self.alignment) != 0:
				raise ValueError(f'Explicitly specified address {addr:#x} must be a multiple of {1 << alignment:#x} bytes')
		else:
			addr = self._align_up(self._next_addr, alignment)

		if not isinstance(size, int) or size < 0:
			raise ValueError(f'Size must be a non-negative integer, not {size!r}')
		size = self._align_up(max(size, 1), alignment)

		if addr > (1 << self.addr_width) or addr + size > (1 << self.addr_width):
			if extend:
				self.addr_width = bits_for(addr + size)
			else:
				raise ValueError(
					f'Address range {addr:#x}..{addr + size:#x} out of bounds for memory map spanning '
					f'range {0:#x}..{1 << self.addr_width:#x} ({self.addr_width} address bits)'
				)

		addr_range = range(addr, addr + size, step)
		overlaps = self._ranges.overlaps(addr_range)
		if overlaps:
			overlap_descrs = []
			for overlap in overlaps:
				if id(overlap) in self._resources:
					_, _, resource_range = self._resources[id(overlap)]
					overlap_descrs.append(f'resource {overlap!r} at {resource_range.start:#x}..{resource_range.stop:#x}')
				if id(overlap) in self._windows:
					_, window_range = self._windows[id(overlap)]
					overlap_descrs.append(f'window {overlap!r} at {window_range.start:#x}..{ window_range.stop:#x}')
			raise ValueError(f'Address range {addr:#x}..{addr + size:#x} overlaps with {", ".join(overlap_descrs)}')

		return addr_range

	def add_resource(
		self, resource: object, *, name: str, size: int, addr: Optional[int] = None,
		alignment: Optional[int] = None, extend: bool = False
	) -> tuple[int, int]:
		'''
		Add a resource.

		A resource is any device on the bus that is a destination for bus transactions, e.g.
		a register or a memory block.

		Arguments
		---------
		resource : object
			Arbitrary object representing a resource.
		name : str
			Name of the resource. It must not collide with the name of other resources or windows
			present in this memory map.
		addr : int or None
			Address of the resource. If ``None``, the implicit next address will be used.
			Otherwise, the exact specified address (which must be a multiple of
			``2 ** max(alignment, self.alignment)``) will be used.
		size : int
			Size of the resource, in minimal addressable units. Rounded up to a multiple of
			``2 ** max(alignment, self.alignment)``.
		alignment : log2 of int or None
			Alignment of the resource. If not specified, the memory map alignment is used.
		extend: bool
			Allow memory map extension. If ``True``, the upper bound of the address space is
			raised as needed, in order to fit a resource that would otherwise be out of bounds.

		Return value
		------------
		A tuple ``(start, end)`` describing the address range assigned to the resource.

		Exceptions
		----------
		Raises :exn:`ValueError` if one of the following occurs:

		- this memory map is frozen;
		- the requested address and size, after alignment, would overlap with any resources or
		windows that have already been added, or would be out of bounds;
		- the resource has already been added to this memory map;
		- the name of the resource is already present in the namespace of this memory map;

		'''

		if self._frozen:
			raise ValueError(f'Memory map has been frozen. Cannot add resource {resource!r}')

		if id(resource) in self._resources:
			_, _, addr_range = self._resources[id(resource)]
			raise ValueError(
				f'Resource {resource!r} is already added at address range '
				f'{addr_range.start:#x}..{addr_range.stop:#x}'
			)

		if not isinstance(name, str) or not name:
			raise TypeError(f'Name must be a non-empty string, not {name!r}')
		if name in self._namespace:
			raise ValueError(f'Name {name} is already used by {self._namespace[name]!r}')

		if alignment is not None:
			if not isinstance(alignment, int) or alignment < 0:
				raise ValueError(f'Alignment must be a non-negative integer, not {alignment!r}')
			alignment = max(alignment, self.alignment)
		else:
			alignment = self.alignment

		addr_range = self._compute_addr_range(addr, size, alignment = alignment, extend = extend)
		self._ranges.insert(addr_range, resource)
		self._resources[id(resource)] = resource, name, addr_range
		self._namespace[name] = resource
		self._next_addr = addr_range.stop
		return (addr_range.start, addr_range.stop)

	def resources(self):
		'''
		Iterate local resources and their address ranges.

		Non-recursively iterate resources in ascending order of their address.

		Yield values
		------------
		A tuple ``resource, name, (start, end)`` describing the address range assigned to the
		resource.

		'''

		for resource, resource_name, resource_range in self._resources.values():
			yield (resource, resource_name, (resource_range.start, resource_range.stop))

	def add_window(
		self, window: 'MemoryMap', *, addr: Optional[int] = None, sparse: Optional[int] = None,
		extend: bool = False
	) -> tuple[int, int, int]:
		'''
		Add a window.

		A window is a device on a bus that provides access to a different bus, i.e. a bus bridge.
		It performs address translation, such that the devices on a subordinate bus have different
		addresses; the memory map reflects this address translation when resources are looked up
		through the window.

		Sparse addressing
		-----------------

		If a narrow bus is bridged to a wide bus, the bridge can perform *sparse* or *dense*
		address translation. In the sparse case, each transaction on the wide bus results in
		one transaction on the narrow bus; high data bits on the wide bus are ignored, and any
		contiguous resource on the narrow bus becomes discontiguous on the wide bus. In the dense
		case, each transaction on the wide bus results in several transactions on the narrow bus,
		and any contiguous resource on the narrow bus stays contiguous on the wide bus.

		Arguments
		---------
		window : :class:`MemoryMap`
			A memory map describing the layout of the window. It is frozen as a side-effect of
			being added to this memory map.
		addr : int or None
			Address of the window. If ``None``, the implicit next address will be used after
			aligning it to ``2 ** window.addr_width``. Otherwise, the exact specified address
			(which must be a multiple of ``2 ** window.addr_width``) will be used.
		sparse : bool or None
			Address translation type. Ignored if the datapath widths of both memory maps are
			equal; must be specified otherwise.
		extend : bool
			Allow memory map extension. If ``True``, the upper bound of the address space is
			raised as needed, in order to fit a window that would otherwise be out of bounds.

		Return value
		------------
		A tuple ``(start, end, ratio)`` describing the address range assigned to the window.
		When bridging buses of unequal data width, ``ratio`` is the amount of contiguous addresses
		on the narrower bus that are accessed for each transaction on the wider bus. Otherwise,
		it is always 1.

		Exceptions
		----------
		Raises :exn:`ValueError` if one of the following occurs:

		- this memory map is frozen;
		- the requested address and size, after alignment, would overlap with any resources or
		windows that have already been added, or would be out of bounds;
		- the added memory map has a wider datapath than this memory map;
		- dense address translation is used and the datapath width of this memory map is not an
		integer multiple of the datapath of the added memory map;
		- the name of the added memory map is already present in the namespace of this memory map;
		- the added memory map has no name, and the name of one of its subordinate resources or
		windows is already present in the namespace of this memory map;

		'''

		if not isinstance(window, MemoryMap):
			raise TypeError(f'Window must be a MemoryMap, not {window!r}')

		if self._frozen:
			raise ValueError(f'Memory map has been frozen. Cannot add window {window!r}')

		if id(window) in self._windows:
			_, addr_range = self._windows[id(window)]
			raise ValueError(f'Window {window!r} is already added at address range {addr_range.start:#x}..{addr_range.stop:#x}')

		if window.data_width > self.data_width:
			raise ValueError(
				f'Window has data width {window.data_width}, and cannot be added to a memory map '
				f'with data width {self.data_width}'
			)
		if window.data_width != self.data_width:
			if sparse is None:
				raise ValueError(
					'Address translation mode must be explicitly specified '
					f'when adding a window with data width {window.data_width} to a memory map '
					f'with data width {self.data_width}'
				)
			if not sparse and self.data_width % window.data_width != 0:
				raise ValueError(
					'Dense addressing cannot be used because the memory map '
					f'data width {self.data_width} is not an integer multiple of window '
					f'data width {window.data_width}'
				)

		if window.name is None:
			name_conflicts = sorted(self._namespace.keys() & window._namespace.keys())
			if name_conflicts:
				name_conflict_descrs = [
					f'{name} is used by {self._namespace[name]!r}' for name in name_conflicts
				]
				raise ValueError(f'The following names are already used: {"; ".join(name_conflict_descrs)}')
		else:
			if window.name in self._namespace:
				raise ValueError(f'Name {window.name} is already used by {self._namespace[window.name]!r}')

		if not sparse:
			ratio = self.data_width // window.data_width
		else:
			ratio = 1
		size = (1 << window.addr_width) // ratio
		# For resources, the alignment argument of add_resource() affects both address and size
		# of the resource; aligning only the address should be done using align_to(). For windows,
		# changing the size (beyond the edge case of the window size being smaller than alignment
		# of the bus) is unlikely to be useful, so there is no alignment argument. The address of
		# a window can still be aligned using align_to().
		alignment = max(self.alignment, window.addr_width // ratio)

		addr_range = self._compute_addr_range(
			addr, size, ratio, alignment = alignment, extend = extend
		)
		window.freeze()
		self._ranges.insert(addr_range, window)
		self._windows[id(window)] = window, addr_range
		if window.name is None:
			self._namespace.update(window._namespace)
		else:
			self._namespace[window.name] = window
		self._next_addr = addr_range.stop
		return addr_range.start, addr_range.stop, addr_range.step

	def windows(self):
		'''
		Iterate local windows and their address ranges.

		Non-recursively iterate windows in ascending order of their address.

		Yield values
		------------
		A tuple ``window, (start, end, ratio)`` describing the address range assigned to
		the window. When bridging buses of unequal data width, ``ratio`` is the amount of
		contiguous addresses on the narrower bus that are accessed for each transaction on
		the wider bus. Otherwise, it is always 1.

		'''

		for window, window_range in self._windows.values():
			yield (window, (window_range.start, window_range.stop, window_range.step))

	def window_patterns(self) -> Generator[tuple[object, tuple[str, int]], None, None]:
		'''
		Iterate local windows and patterns that match their address ranges.

		Non-recursively iterate windows in ascending order of their address.

		Yield values
		------------
		A tuple ``window, (pattern, ratio)`` describing the address range assigned to the window.
		``pattern`` is a ``self.addr_width`` wide pattern that may be used in ``Case`` or ``match``
		to determine if an address signal is within the address range of ``window``. When bridging
		buses of unequal data width, ``ratio`` is the amount of contiguous addresses on
		the narrower bus that are accessed for each transaction on the wider bus. Otherwise,
		it is always 1.

		'''

		for window, window_range in self._windows.values():
			const_bits = self.addr_width - window.addr_width
			if const_bits > 0:
				const_pat = f'{window_range.start >> window.addr_width:0{const_bits}b}'
			else:
				const_pat = ''
			pattern = f'{const_pat}{"-" * window.addr_width}'
			yield (window, (pattern, window_range.step))

	@staticmethod
	def _translate(
		resource_info: ResourceInfo, window, window_range: range
	) -> ResourceInfo:
		assert (resource_info.end - resource_info.start) % window_range.step == 0
		# Accessing a resource through a dense and then a sparse window results in very strange
		# layouts that cannot be easily represented, so reject those.
		assert window_range.step == 1 or resource_info.width == window.data_width

		name  = resource_info.name if window.name is None else (window.name, *resource_info.name)
		size  = (resource_info.end - resource_info.start) // window_range.step
		start = resource_info.start + window_range.start
		width = resource_info.width * window_range.step
		return ResourceInfo(resource_info.resource, name, start, start + size, width)

	def all_resources(self) -> Generator[ResourceInfo, None, None]:
		'''
		Iterate all resources and their address ranges.

		Recursively iterate all resources in ascending order of their address, performing address
		translation for resources that are located behind a window.

		Yield values
		------------
		An instance of :class:`ResourceInfo` describing the resource and its address range.

		'''

		for addr_range, assignment in self._ranges.items():
			if id(assignment) in self._resources:
				_, resource_name, _ = self._resources[id(assignment)]
				yield ResourceInfo(
					assignment, resource_name, addr_range.start, addr_range.stop, self.data_width
				)
			elif id(assignment) in self._windows:
				for resource_info in assignment.all_resources():
					yield self._translate(resource_info, assignment, addr_range)
			else:
				assert False # :nocov:

	def find_resource(self, resource: object) -> ResourceInfo:
		'''
		Find address range corresponding to a resource.

		Recursively find the address range of a resource, performing address translation for
		resources that are located behind a window.

		Arguments
		---------
		resource
			Resource previously added to this memory map or any windows.

		Return value
		------------
		An instance of :class:`ResourceInfo` describing the resource and its address range.

		Exceptions
		----------
		Raises :exn:`KeyError` if the resource is not found.

		'''

		if id(resource) in self._resources:
			_, resource_name, resource_range = self._resources[id(resource)]
			return ResourceInfo(resource, resource_name, resource_range.start, resource_range.stop,
								self.data_width)

		for window, window_range in self._windows.values():
			try:
				return self._translate(window.find_resource(resource), window, window_range)
			except KeyError:
				pass

		raise KeyError(resource)

	def decode_address(self, address: int) -> Optional[object]:
		'''
		Decode an address to a resource.

		Arguments
		---------
		address : int
			Address of interest.

		Return value
		------------
		A resource mapped to the provided address, or ``None`` if there is no such resource.

		'''

		assignment = self._ranges.get(address)
		if assignment is None:
			return

		if id(assignment) in self._resources:
			return assignment
		elif id(assignment) in self._windows:
			_, addr_range = self._windows[id(assignment)]
			return assignment.decode_address((address - addr_range.start) // addr_range.step)
		else:
			assert False # :nocov:



# SPDX-License-Identifier: BSD-2-Clause

from sys      import _getframe, excepthook
from types    import TracebackType
from typing   import Optional, Type
from warnings import warn_explicit

from ..util   import get_linter_option

__all__ = (
	'MustUse',
	'UnusedMustUse',
)

class UnusedMustUse(Warning):
	pass


class MustUse:
	_MustUse__silence = False
	_MustUse__warning = UnusedMustUse

	def __new__(cls, *args, src_loc_at: int = 0, **kwargs) -> 'MustUse':
		frame = _getframe(1 + src_loc_at)
		self = super().__new__(cls)
		self._MustUse__used    = False
		self._MustUse__context = dict(
			filename = frame.f_code.co_filename,
			lineno   = frame.f_lineno,
			source   = self
		)
		return self

	def __del__(self) -> None:
		if self._MustUse__silence:
			return
		if getattr(self._MustUse__warning, '_MustUse__silence', False):
			return
		if hasattr(self, '_MustUse__used') and not self._MustUse__used:
			if get_linter_option(
				self._MustUse__context['filename'],
				self._MustUse__warning.__name__, bool, True
			):
				warn_explicit(
					f'{self!r} created but never used',
					self._MustUse__warning,
					**self._MustUse__context
				)

_old_excepthook = excepthook
def _silence_elaboratable(
	type: Type[BaseException], value: BaseException, traceback: Optional[TracebackType]
) -> None:
	# Don't show anything if the interpreter crashed; that'd just obscure the exception
	# traceback instead of helping.
	MustUse._MustUse__silence = True
	_old_excepthook(type, value, traceback)
excepthook = _silence_elaboratable



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from collections   import OrderedDict
from enum          import Enum
from sys           import version_info

from torii.hdl.ast import *
from torii.hdl.cd  import *
from torii.hdl.dsl import *

from ..utils       import ToriiTestSuiteCase

class DSLTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.s1 = Signal()
		self.s2 = Signal()
		self.s3 = Signal()
		self.c1 = Signal()
		self.c2 = Signal()
		self.c3 = Signal()
		self.w1 = Signal(4)

	def test_cant_inherit(self):
		with self.assertRaisesRegex(
			SyntaxError, (
				r'^Instead of inheriting from `Module`, inherit from `Elaboratable` and '
				r'return a `Module` from the `elaborate\(self, platform\)` method$'
			)
		):
			class ORGate(Module):
				pass

	def test_d_comb(self):
		m = Module()
		m.d.comb += self.c1.eq(1)
		m._flush()
		self.assertEqual(m._driving[self.c1], None)
		self.assertRepr(m._statements, '''(
			(eq (sig c1) (const 1'd1))
		)''')

	def test_d_sync(self):
		m = Module()
		m.d.sync += self.c1.eq(1)
		m._flush()
		self.assertEqual(m._driving[self.c1], 'sync')
		self.assertRepr(m._statements, '''(
			(eq (sig c1) (const 1'd1))
		)''')

	def test_d_pix(self):
		m = Module()
		m.d.pix += self.c1.eq(1)
		m._flush()
		self.assertEqual(m._driving[self.c1], 'pix')
		self.assertRepr(m._statements, '''(
			(eq (sig c1) (const 1'd1))
		)''')

	def test_d_index(self):
		m = Module()
		m.d['pix'] += self.c1.eq(1)
		m._flush()
		self.assertEqual(m._driving[self.c1], 'pix')
		self.assertRepr(m._statements, '''(
			(eq (sig c1) (const 1'd1))
		)''')

	def test_d_no_conflict(self):
		m = Module()
		m.d.comb += self.w1[0].eq(1)
		m.d.comb += self.w1[1].eq(1)

	def test_d_conflict(self):
		m = Module()
		with self.assertRaisesRegex(
			SyntaxError, (
				r'^Driver-driver conflict: trying to drive \(sig c1\) from d\.sync, but it '
				r'is already driven from d\.comb$'
			)
		):
			m.d.comb += self.c1.eq(1)
			m.d.sync += self.c1.eq(1)

	def test_d_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			AttributeError,
			r'^Cannot assign \'d\.pix\' attribute; did you mean \'d.pix \+=\'\?$'
		):
			m.d.pix = None

	def test_d_asgn_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			SyntaxError,
			r'^Only assignments and property checks may be appended to d\.sync$'
		):
			m.d.sync += Switch(self.s1, {})

	def test_comb_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			AttributeError,
			r'^\'Module\' object has no attribute \'comb\'; did you mean \'d\.comb\'\?$'
		):
			m.comb += self.c1.eq(1)

	def test_sync_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			AttributeError,
			r'^\'Module\' object has no attribute \'sync\'; did you mean \'d\.sync\'\?$'
		):
			m.sync += self.c1.eq(1)

	def test_attr_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			AttributeError,
			r'^\'Module\' object has no attribute \'nonexistentattr\'$'
		):
			m.nonexistentattr

	def test_d_suspicious(self):
		m = Module()
		with self.assertWarnsRegex(
			SyntaxWarning, (
				r'^Using \'<module>\.d\.submodules\' would add statements to clock domain '
				r'\'submodules\'; did you mean <module>\.submodules instead\?$'
			)
		):
			m.d.submodules += []

	def test_clock_signal(self):
		m = Module()
		m.d.comb += ClockSignal('pix').eq(ClockSignal())
		self.assertRepr(m._statements, '''
		(
			(eq (clk pix) (clk sync))
		)
		''')

	def test_reset_signal(self):
		m = Module()
		m.d.comb += ResetSignal('pix').eq(1)
		self.assertRepr(m._statements, '''
		(
			(eq (rst pix) (const 1'd1))
		)
		''')

	def test_sample_domain(self):
		m = Module()
		i = Signal()
		o1 = Signal()
		o2 = Signal()
		o3 = Signal()
		m.d.sync += o1.eq(Past(i))
		m.d.pix  += o2.eq(Past(i))
		m.d.pix  += o3.eq(Past(i, domain = 'sync'))
		f = m.elaborate(platform = None)
		self.assertRepr(f.statements, '''
		(
			(eq (sig o1) (sample (sig i) @ sync[1]))
			(eq (sig o2) (sample (sig i) @ pix[1]))
			(eq (sig o3) (sample (sig i) @ sync[1]))
		)
		''')

	def test_If(self):
		m = Module()
		with m.If(self.s1):
			m.d.comb += self.c1.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (sig s1))
				(case 1 (eq (sig c1) (const 1'd1)))
			)
		)
		''')

	def test_If_Elif(self):
		m = Module()
		with m.If(self.s1):
			m.d.comb += self.c1.eq(1)
		with m.Elif(self.s2):
			m.d.sync += self.c2.eq(0)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (sig s1) (sig s2))
				(case -1 (eq (sig c1) (const 1'd1)))
				(case 1- (eq (sig c2) (const 1'd0)))
			)
		)
		''')

	def test_If_Elif_Else(self):
		m = Module()
		with m.If(self.s1):
			m.d.comb += self.c1.eq(1)
		with m.Elif(self.s2):
			m.d.sync += self.c2.eq(0)
		with m.Else():
			m.d.comb += self.c3.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (sig s1) (sig s2))
				(case -1 (eq (sig c1) (const 1'd1)))
				(case 1- (eq (sig c2) (const 1'd0)))
				(default (eq (sig c3) (const 1'd1)))
			)
		)
		''')

	def test_If_If(self):
		m = Module()
		with m.If(self.s1):
			m.d.comb += self.c1.eq(1)
		with m.If(self.s2):
			m.d.comb += self.c2.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (sig s1))
				(case 1 (eq (sig c1) (const 1'd1)))
			)
			(switch (cat (sig s2))
				(case 1 (eq (sig c2) (const 1'd1)))
			)
		)
		''')

	def test_If_nested_If(self):
		m = Module()
		with m.If(self.s1):
			m.d.comb += self.c1.eq(1)
			with m.If(self.s2):
				m.d.comb += self.c2.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (sig s1))
				(case 1 (eq (sig c1) (const 1'd1))
					(switch (cat (sig s2))
						(case 1 (eq (sig c2) (const 1'd1)))
					)
				)
			)
		)
		''')

	def test_If_dangling_Else(self):
		m = Module()
		with m.If(self.s1):
			m.d.comb += self.c1.eq(1)
			with m.If(self.s2):
				m.d.comb += self.c2.eq(1)
		with m.Else():
			m.d.comb += self.c3.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (sig s1))
				(case 1
					(eq (sig c1) (const 1'd1))
					(switch (cat (sig s2))
						(case 1 (eq (sig c2) (const 1'd1)))
					)
				)
				(default
					(eq (sig c3) (const 1'd1))
				)
			)
		)
		''')

	def test_Elif_wrong(self):
		m = Module()
		with self.assertRaisesRegex(SyntaxError, r'^Elif without preceding If$'):
			with m.Elif(self.s2):
				pass

	def test_Elif_wrong_nested(self):
		m = Module()
		with m.If(self.s1):
			with self.assertRaisesRegex(SyntaxError, r'^Elif without preceding If$'):
				with m.Elif(self.s2):
					pass

	def test_Else_wrong(self):
		m = Module()
		with self.assertRaisesRegex(SyntaxError, r'^Else without preceding If\/Elif$'):
			with m.Else():
				pass

	def test_Else_wrong_nested(self):
		m = Module()
		with m.If(self.s1):
			with self.assertRaisesRegex(SyntaxError, r'^Else without preceding If/Elif$'):
				with m.Else():
					pass

	def test_Elif_Elif_wrong_nested(self):
		m = Module()
		with m.If(self.s1):
			pass
		with m.Elif(self.s2):
			with self.assertRaisesRegex(SyntaxError, r'^Elif without preceding If$'):
				with m.Elif(self.s3):
					pass

	def test_Else_Else_wrong_nested(self):
		m = Module()
		with m.If(self.s1):
			pass
		with m.Else():
			with self.assertRaisesRegex(SyntaxError, r'^Else without preceding If/Elif$'):
				with m.Else():
					pass

	def test_If_wide(self):
		m = Module()
		with m.If(self.w1):
			m.d.comb += self.c1.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (cat (b (sig w1)))
				(case 1 (eq (sig c1) (const 1'd1)))
			)
		)
		''')

	if version_info < (3, 12):
		def test_If_signed_suspicious(self):
			m = Module()
			with self.assertWarnsRegex(
				SyntaxWarning, (
					r'^Signed values in If\/Elif conditions usually result from inverting Python '
					r'booleans with ~, which leads to unexpected results\. Replace `~flag` with '
					r'`not flag`\. \(If this is a false positive, silence this warning with '
					r'`m\.If\(x\)` → `m\.If\(x\.bool\(\)\)`\.\)$'
				)
			):
				with m.If(~True):
					pass

		def test_Elif_signed_suspicious(self):
			m = Module()
			with m.If(0):
				pass
			with self.assertWarnsRegex(
				SyntaxWarning, (
					r'^Signed values in If\/Elif conditions usually result from inverting Python '
					r'booleans with ~, which leads to unexpected results\. Replace `~flag` with '
					r'`not flag`\. \(If this is a false positive, silence this warning with '
					r'`m\.If\(x\)` → `m\.If\(x\.bool\(\)\)`\.\)$'
				)
			):
				with m.Elif(~True):
					pass

	def test_if_If_Elif_Else(self):
		m = Module()
		with self.assertRaisesRegex(
			SyntaxError,
			r'^`if m\.If\(\.\.\.\):` does not work; use `with m\.If\(\.\.\.\)`$'
		):
			if m.If(0):
				pass
		with m.If(0):
			pass
		with self.assertRaisesRegex(
			SyntaxError,
			r'^`if m\.Elif\(\.\.\.\):` does not work; use `with m\.Elif\(\.\.\.\)`$'
		):
			if m.Elif(0):
				pass
		with self.assertRaisesRegex(
			SyntaxError,
			r'^`if m\.Else\(\.\.\.\):` does not work; use `with m\.Else\(\.\.\.\)`$'
		):
			if m.Else():
				pass

	def test_Switch(self):
		m = Module()
		with m.Switch(self.w1):
			with m.Case(3):
				m.d.comb += self.c1.eq(1)
			with m.Case('11--'):
				m.d.comb += self.c2.eq(1)
			with m.Case('1 0--'):
				m.d.comb += self.c2.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (sig w1)
				(case 0011 (eq (sig c1) (const 1'd1)))
				(case 11-- (eq (sig c2) (const 1'd1)))
				(case 10-- (eq (sig c2) (const 1'd1)))
			)
		)
		''')

	def test_Switch_default_Case(self):
		m = Module()
		with m.Switch(self.w1):
			with m.Case(3):
				m.d.comb += self.c1.eq(1)
			with m.Case():
				m.d.comb += self.c2.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (sig w1)
				(case 0011 (eq (sig c1) (const 1'd1)))
				(default (eq (sig c2) (const 1'd1)))
			)
		)
		''')

	def test_Switch_default_Default(self):
		m = Module()
		with m.Switch(self.w1):
			with m.Case(3):
				m.d.comb += self.c1.eq(1)
			with m.Default():
				m.d.comb += self.c2.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (sig w1)
				(case 0011 (eq (sig c1) (const 1'd1)))
				(default (eq (sig c2) (const 1'd1)))
			)
		)
		''')

	def test_Switch_const_test(self):
		m = Module()
		with m.Switch(1):
			with m.Case(1):
				m.d.comb += self.c1.eq(1)
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (const 1'd1)
				(case 1 (eq (sig c1) (const 1'd1)))
			)
		)
		''')

	def test_Switch_enum(self):
		class Color(Enum):
			RED  = 1
			BLUE = 2
		m = Module()
		se = Signal(Color)
		with m.Switch(se):
			with m.Case(Color.RED):
				m.d.comb += self.c1.eq(1)
		self.assertRepr(m._statements, '''
		(
			(switch (sig se)
				(case 01 (eq (sig c1) (const 1'd1)))
			)
		)
		''')

	def test_Case_width_wrong(self):
		class Color(Enum):
			RED = 0b10101010
		m = Module()
		with m.Switch(self.w1):
			with self.assertRaisesRegex(
				SyntaxError,
				r'^Case pattern \'--\' must have the same width as switch value \(which is 4\)$'
			):
				with m.Case('--'):
					pass
			with self.assertWarnsRegex(
				SyntaxWarning, (
					r'^Case pattern \'10110\' is wider than switch value \(which has width 4\); '
					r'comparison will never be true$'
				)
			):
				with m.Case(0b10110):
					pass
			with self.assertWarnsRegex(
				SyntaxWarning, (
					r'^Case pattern \'10101010\' \(Color\.RED\) is wider than switch value '
					r'\(which has width 4\); comparison will never be true$'
				)
			):
				with m.Case(Color.RED):
					pass
		self.assertRepr(m._statements, '''
		(
			(switch (sig w1) )
		)
		''')

	def test_Case_bits_wrong(self):
		m = Module()
		with m.Switch(self.w1):
			with self.assertRaisesRegex(
				SyntaxError, (
					r'^Case pattern \'abc\' must consist of 0, 1, and - \(don\'t care\) bits, '
					r'and may include whitespace$'
				)
			):
				with m.Case('abc'):
					pass

	def test_Case_pattern_wrong(self):
		m = Module()
		with m.Switch(self.w1):
			with self.assertRaisesRegex(
				SyntaxError,
				r'^Case pattern must be an integer, a string, or an enumeration, not 1\.0$'
			):
				with m.Case(1.0):
					pass

	def test_Case_outside_Switch_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			SyntaxError,
			r'^Case is not permitted outside of Switch$'
		):
			with m.Case():
				pass

	def test_If_inside_Switch_wrong(self):
		m = Module()
		with m.Switch(self.s1):
			with self.assertRaisesRegex(
				SyntaxError, (
					r'^If is not permitted directly inside of Switch; '
					r'it is permitted inside of Switch Case$'
				)
			):
				with m.If(self.s2):
					pass

	def test_Case_wrong_nested(self):
		m = Module()
		with m.Switch(self.s1):
			with m.Case(0):
				with self.assertRaisesRegex(
					SyntaxError,
					r'^Case is not permitted outside of Switch$'
				):
					with m.Case(1):
						pass

	def test_FSM_basic(self):
		a = Signal()
		b = Signal()
		c = Signal()
		m = Module()
		with m.FSM():
			with m.State('FIRST'):
				m.d.comb += a.eq(1)
				m.next = 'SECOND'
			with m.State('SECOND'):
				m.d.sync += b.eq(~b)
				with m.If(c):
					m.next = 'FIRST'
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (sig fsm_state)
				(case 0
					(eq (sig a) (const 1'd1))
					(eq (sig fsm_state) (const 1'd1))
				)
				(case 1
					(eq (sig b) (~ (sig b)))
					(switch (cat (sig c))
						(case 1
							(eq (sig fsm_state) (const 1'd0)))
					)
				)
			)
		)
		''')
		self.assertEqual({repr(k): v for k, v in m._driving.items()}, {
			'(sig a)': None,
			'(sig fsm_state)': 'sync',
			'(sig b)': 'sync',
		})

		frag = m.elaborate(platform = None)
		fsm  = frag.find_generated('fsm')
		self.assertIsInstance(fsm.state, Signal)
		self.assertEqual(fsm.encoding, OrderedDict({
			'FIRST': 0,
			'SECOND': 1,
		}))
		self.assertEqual(fsm.decoding, OrderedDict({
			0: 'FIRST',
			1: 'SECOND'
		}))

	def test_FSM_reset(self):
		a = Signal()
		m = Module()
		with m.FSM(reset = 'SECOND'):
			with m.State('FIRST'):
				m.d.comb += a.eq(0)
				m.next = 'SECOND'
			with m.State('SECOND'):
				m.next = 'FIRST'
		m._flush()
		self.assertRepr(m._statements, '''
		(
			(switch (sig fsm_state)
				(case 0
					(eq (sig a) (const 1'd0))
					(eq (sig fsm_state) (const 1'd1))
				)
				(case 1
					(eq (sig fsm_state) (const 1'd0))
				)
			)
		)
		''')

	def test_FSM_ongoing(self):
		a = Signal()
		b = Signal()
		m = Module()
		with m.FSM() as fsm:
			m.d.comb += b.eq(fsm.ongoing('SECOND'))
			with m.State('FIRST'):
				pass
			m.d.comb += a.eq(fsm.ongoing('FIRST'))
			with m.State('SECOND'):
				pass
		m._flush()
		self.assertEqual(m._generated['fsm'].state.reset, 1)
		self.maxDiff = 10000
		self.assertRepr(m._statements, '''
		(
			(eq (sig b) (== (sig fsm_state) (const 1'd0)))
			(eq (sig a) (== (sig fsm_state) (const 1'd1)))
			(switch (sig fsm_state)
				(case 1
				)
				(case 0
				)
			)
		)
		''')

	def test_FSM_empty(self):
		m = Module()
		with m.FSM():
			pass
		self.assertRepr(m._statements, '''
		()
		''')

	def test_FSM_wrong_domain(self):
		m = Module()
		with self.assertRaisesRegex(
			ValueError,
				r'^FSM may not be driven by the \'comb\' domain$'
		):
			with m.FSM(domain = 'comb'):
				pass

	def test_FSM_wrong_undefined(self):
		m = Module()
		with self.assertRaisesRegex(
			NameError,
			r'^FSM state \'FOO\' is referenced but not defined$'
		):
			with m.FSM() as fsm:
				fsm.ongoing('FOO')

	def test_FSM_wrong_redefined(self):
		m = Module()
		with m.FSM():
			with m.State('FOO'):
				pass
			with self.assertRaisesRegex(
				NameError,
				r'^FSM state \'FOO\' is already defined$'
			):
				with m.State('FOO'):
					pass

	def test_FSM_wrong_next(self):
		m = Module()
		with self.assertRaisesRegex(
			SyntaxError,
			r'^Only assignment to `m\.next` is permitted$'
		):
			m.next
		with self.assertRaisesRegex(
			SyntaxError,
			r'^`m\.next = <\.\.\.>` is only permitted inside an FSM state$'
		):
			m.next = 'FOO'
		with self.assertRaisesRegex(
			SyntaxError,
			r'^`m\.next = <\.\.\.>` is only permitted inside an FSM state$'
		):
			with m.FSM():
				m.next = 'FOO'

	def test_If_inside_FSM_wrong(self):
		m = Module()
		with m.FSM():
			with m.State('FOO'):
				pass
			with self.assertRaisesRegex(
				SyntaxError, (
					r'^If is not permitted directly inside of FSM; '
					r'it is permitted inside of FSM State$'
				)
			):
				with m.If(self.s2):
					pass

	def test_State_outside_FSM_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			SyntaxError,
			r'^FSM State is not permitted outside of FSM'
		):
			with m.State('FOO'):
				pass


	def test_FSM_State_wrong_nested(self):
		m = Module()
		with m.FSM():
			with m.State('FOO'):
				with self.assertRaisesRegex(
					SyntaxError,
					r'^FSM State is not permitted outside of FSM'
				):
					with m.State('BAR'):
						pass

	def test_auto_pop_ctrl(self):
		m = Module()
		with m.If(self.w1):
			m.d.comb += self.c1.eq(1)
		m.d.comb += self.c2.eq(1)
		self.assertRepr(m._statements, '''
		(
			(switch (cat (b (sig w1)))
				(case 1 (eq (sig c1) (const 1'd1)))
			)
			(eq (sig c2) (const 1'd1))
		)
		''')

	def test_submodule_anon(self):
		m1 = Module()
		m2 = Module()
		m1.submodules += m2
		self.assertEqual(m1._anon_submodules, [m2])
		self.assertEqual(m1._named_submodules, {})

	def test_submodule_anon_multi(self):
		m1 = Module()
		m2 = Module()
		m3 = Module()
		m1.submodules += m2, m3
		self.assertEqual(m1._anon_submodules, [m2, m3])
		self.assertEqual(m1._named_submodules, {})

	def test_submodule_named(self):
		m1 = Module()
		m2 = Module()
		m1.submodules.foo = m2
		self.assertEqual(m1._anon_submodules, [])
		self.assertEqual(m1._named_submodules, {'foo': m2})

	def test_submodule_named_index(self):
		m1 = Module()
		m2 = Module()
		m1.submodules['foo'] = m2
		self.assertEqual(m1._anon_submodules, [])
		self.assertEqual(m1._named_submodules, {'foo': m2})

	def test_submodule_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			TypeError,
			r'^Trying to add 1, which does not implement \.elaborate\(\), as a submodule$'
		):
			m.submodules.foo = 1
		with self.assertRaisesRegex(
			TypeError,
			r'^Trying to add 1, which does not implement \.elaborate\(\), as a submodule$'
		):
			m.submodules += 1

	def test_submodule_named_conflict(self):
		m1 = Module()
		m2 = Module()
		m1.submodules.foo = m2
		with self.assertRaisesRegex(NameError, r'^Submodule named \'foo\' already exists$'):
			m1.submodules.foo = m2

	def test_submodule_get(self):
		m1 = Module()
		m2 = Module()
		m1.submodules.foo = m2
		m3 = m1.submodules.foo
		self.assertEqual(m2, m3)

	def test_submodule_get_index(self):
		m1 = Module()
		m2 = Module()
		m1.submodules['foo'] = m2
		m3 = m1.submodules['foo']
		self.assertEqual(m2, m3)

	def test_submodule_get_unset(self):
		m1 = Module()
		with self.assertRaisesRegex(AttributeError, r'^No submodule named \'foo\' exists$'):
			m1.submodules.foo
		with self.assertRaisesRegex(AttributeError, r'^No submodule named \'foo\' exists$'):
			m1.submodules['foo']

	def test_domain_named_implicit(self):
		m = Module()
		m.domains += ClockDomain('sync')
		self.assertEqual(len(m._domains), 1)

	def test_domain_named_explicit(self):
		m = Module()
		m.domains.foo = ClockDomain()
		self.assertEqual(len(m._domains), 1)
		self.assertEqual(m._domains['foo'].name, 'foo')

	def test_domain_add_wrong(self):
		m = Module()
		with self.assertRaisesRegex(
			TypeError,
			r'^Only clock domains may be added to `m\.domains`, not 1$'
		):
			m.domains.foo = 1
		with self.assertRaisesRegex(
			TypeError,
			r'^Only clock domains may be added to `m\.domains`, not 1$'
		):
			m.domains += 1

	def test_domain_add_wrong_name(self):
		m = Module()
		with self.assertRaisesRegex(
			NameError,
			r'^Clock domain name \'bar\' must match name in `m\.domains\.foo \+= \.\.\.` syntax$'
		):
			m.domains.foo = ClockDomain('bar')

	def test_domain_add_wrong_duplicate(self):
		m = Module()
		m.domains += ClockDomain('foo')
		with self.assertRaisesRegex(
			NameError,
			r'^Clock domain named \'foo\' already exists$'
		):
			m.domains += ClockDomain('foo')

	def test_lower(self):
		m1 = Module()
		m1.d.comb += self.c1.eq(self.s1)
		m2 = Module()
		m2.d.comb += self.c2.eq(self.s2)
		m2.d.sync += self.c3.eq(self.s3)
		m1.submodules.foo = m2

		f1 = m1.elaborate(platform = None)
		self.assertRepr(f1.statements, '''
		(
			(eq (sig c1) (sig s1))
		)
		''')
		self.assertEqual(f1.drivers, {
			None: SignalSet((self.c1,))
		})
		self.assertEqual(len(f1.subfragments), 1)
		(f2, f2_name), = f1.subfragments
		self.assertEqual(f2_name, 'foo')
		self.assertRepr(f2.statements, '''
		(
			(eq (sig c2) (sig s2))
			(eq (sig c3) (sig s3))
		)
		''')
		self.assertEqual(f2.drivers, {
			None: SignalSet((self.c2,)),
			'sync': SignalSet((self.c3,))
		})
		self.assertEqual(len(f2.subfragments), 0)



# SPDX-License-Identifier: BSD-2-Clause

from torii.util import string as util_string
from torii.test import ToriiTestCase

class StringUtilTestCase(ToriiTestCase):

	def test_ascii_escape(self):
		self.assertEqual(util_string.ascii_escape('abcdefg12346_'), 'abcdefg12346_')
		self.assertEqual(
			util_string.ascii_escape('-!@#$%^&*()'),
			''.join(map(lambda c: f'_{ord(c):02x}_', '-!@#$%^&*()'))
		)
		self.assertEqual(util_string.ascii_escape('nya~'), 'nya_7e_')

	def test_tcl_escape(self):
		self.assertEqual(util_string.tcl_escape('uwu'), '{uwu}')
		self.assertEqual(util_string.tcl_escape('"n"{y}a'), '{"n"\\{y\\}a}')
		self.assertEqual(
			util_string.tcl_escape(r'(fo"o\)B"ar[B{a}Z]'),
			'{(fo"o\\\\)B"ar[B\\{a\\}Z]}'
		)

	def test_tcl_quote(self):
		self.assertEqual(util_string.tcl_quote('test'), '"test"')
		self.assertEqual(util_string.tcl_quote('test"123'), '"test\\"123"')
		self.assertEqual(util_string.tcl_quote('test[123'), '"test\\[123"')
		self.assertEqual(util_string.tcl_quote('test$123'), '"test\\$123"')
		self.assertEqual(util_string.tcl_quote('"test$123['), '"\\"test\\$123\\["')



# SPDX-License-Identifier: BSD-2-Clause

from torii.lib.soc.memory import _RangeMap, ResourceInfo, MemoryMap

from ...utils             import ToriiTestSuiteCase

class RangeMapTestCase(ToriiTestSuiteCase):
	def test_insert(self):
		range_map = _RangeMap()
		range_map.insert(range(0, 10), 'a')
		range_map.insert(range(20, 21), 'c')
		range_map.insert(range(15, 16), 'b')
		range_map.insert(range(16, 20), 'q')
		self.assertEqual(range_map._keys, [
			range(0, 10), range(15, 16), range(16, 20), range(20, 21)
		])

	def test_overlaps(self):
		range_map = _RangeMap()
		range_map.insert(range(10, 20), 'a')
		self.assertEqual(range_map.overlaps(range(5, 15)), ['a'])
		self.assertEqual(range_map.overlaps(range(15, 25)), ['a'])
		self.assertEqual(range_map.overlaps(range(5, 25)), ['a'])
		self.assertEqual(range_map.overlaps(range(0, 3)), [])
		self.assertEqual(range_map.overlaps(range(0, 5)), [])
		self.assertEqual(range_map.overlaps(range(25, 30)), [])

	def test_insert_wrong_overlap(self):
		range_map = _RangeMap()
		range_map.insert(range(0, 10), 'a')
		with self.assertRaises(AssertionError):
			range_map.insert(range(5, 15), 'b')

	def test_get(self):
		range_map = _RangeMap()
		range_map.insert(range(5, 15), 'a')
		self.assertEqual(range_map.get(0), None)
		self.assertEqual(range_map.get(5), 'a')
		self.assertEqual(range_map.get(10), 'a')
		self.assertEqual(range_map.get(14), 'a')
		self.assertEqual(range_map.get(15), None)


class ResourceInfoTestCase(ToriiTestSuiteCase):
	def test_simple(self):
		info = ResourceInfo('a', name = ('foo', 'bar'), start = 0, end = 1, width = 8)
		self.assertEqual(info.name, ('foo', 'bar'))
		self.assertEqual(info.start, 0)
		self.assertEqual(info.end, 1)
		self.assertEqual(info.width, 8)

	def test_name_cast(self):
		info = ResourceInfo('a', name = 'foo', start = 0, end = 1, width = 8)
		self.assertEqual(info.name, ('foo',))

	def test_wrong_name(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Name must be a non-empty sequence of non-empty strings, not \(1,\)'
		):
			ResourceInfo('a', name = (1,), start = 0, end = 1, width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Name must be a non-empty sequence of non-empty strings, not \(\)'
		):
			ResourceInfo('a', name = (), start = 0, end = 1, width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Name must be a non-empty sequence of non-empty strings, not \(\'foo\', \'\'\)'
		):
			ResourceInfo('a', name = ('foo', ''), start = 0, end = 1, width = 8)

	def test_wrong_start_addr(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Start address must be a non-negative integer, not \'foo\''
		):
			ResourceInfo('a', name = 'b', start = 'foo', end = 1, width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Start address must be a non-negative integer, not -1'
		):
			ResourceInfo('a', name = 'b', start = -1, end = 1, width = 8)

	def test_wrong_end_addr(self):
		with self.assertRaisesRegex(
			TypeError,
			r'End address must be an integer greater than the start address, not \'foo\''
		):
			ResourceInfo('a', name = 'b', start = 0, end = 'foo', width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'End address must be an integer greater than the start address, not 0'
		):
			ResourceInfo('a', name = 'b', start = 0, end = 0, width = 8)

	def test_wrong_width(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Width must be a non-negative integer, not \'foo\''
		):
			ResourceInfo('a', name = 'b', start = 0, end = 1, width = 'foo')
		with self.assertRaisesRegex(
			TypeError,
			r'Width must be a non-negative integer, not -1'
		):
			ResourceInfo('a', name = 'b', start = 0, end = 1, width = -1)


class MemoryMapTestCase(ToriiTestSuiteCase):
	def test_name(self):
		memory_map_0 = MemoryMap(addr_width = 1, data_width = 8)
		memory_map_1 = MemoryMap(addr_width = 1, data_width = 8, name = None)
		memory_map_2 = MemoryMap(addr_width = 1, data_width = 8, name = 'foo')
		self.assertEqual(memory_map_0.name, None)
		self.assertEqual(memory_map_1.name, None)
		self.assertEqual(memory_map_2.name, 'foo')

	def test_wrong_name(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Name must be a non-empty string, not 1'
		):
			MemoryMap(addr_width = 1, data_width = 8, name = 1)
		with self.assertRaisesRegex(ValueError,
				r'Name must be a non-empty string, not \'\''
		):
			MemoryMap(addr_width = 1, data_width = 8, name = '')

	def test_wrong_addr_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Address width must be a positive integer, not -1'
		):
			MemoryMap(addr_width = -1, data_width = 8)

	def test_wrong_data_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Data width must be a positive integer, not -1'
		):
			MemoryMap(addr_width = 16, data_width = -1)

	def test_wrong_alignment(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Alignment must be a non-negative integer, not -1'
		):
			MemoryMap(addr_width = 16, data_width = 8, alignment = -1)

	def test_set_addr_width_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Address width must be a positive integer, not -1'
		):
			memory_map = MemoryMap(addr_width = 1, data_width = 8)
			memory_map.addr_width = -1

	def test_set_addr_width_wrong_shrink(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Address width 1 must not be less than its previous value 2, '
			r'because resources that were previously added may not fit anymore'
		):
			memory_map = MemoryMap(addr_width = 2, data_width = 8)
			memory_map.addr_width = 1

	def test_set_addr_width_wrong_frozen(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has been frozen. Address width cannot be extended '
			r'further'
		):
			memory_map = MemoryMap(addr_width = 1, data_width = 8)
			memory_map.freeze()
			memory_map.addr_width = 2

	def test_add_resource(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 1), (0, 1))
		self.assertEqual(memory_map.add_resource(resource = 'b', name = 'bar', size = 2), (1, 3))

	def test_add_resource_map_aligned(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8, alignment = 1)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 1), (0, 2))
		self.assertEqual(memory_map.add_resource('b', name = 'bar', size = 2), (2, 4))

	def test_add_resource_explicit_aligned(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 1), (0, 1))
		self.assertEqual(memory_map.add_resource('b', name = 'bar', size = 1, alignment = 1), (2, 4))
		self.assertEqual(memory_map.add_resource('c', name = 'baz', size = 2), (4, 6))

	def test_add_resource_addr(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 1, addr = 10), (10, 11))
		self.assertEqual(memory_map.add_resource('b', name = 'bar', size = 2), (11, 13))

	def test_add_resource_extend(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(
			memory_map.add_resource(
				'a', name = 'foo', size = 1, addr = 0x10000, extend = True
			),
			(0x10000, 0x10001)
		)
		self.assertEqual(memory_map.addr_width, 17)

	def test_add_resource_size_zero(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 0), (0, 1))
		self.assertEqual(memory_map.add_resource('b', name = 'bar', size = 0), (1, 2))

	def test_add_resource_wrong_frozen(self):
		memory_map = MemoryMap(addr_width = 2, data_width = 8)
		memory_map.freeze()
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has been frozen. Cannot add resource \'a\''
		):
			memory_map.add_resource('a', name = 'foo', size = 0)

	def test_add_resource_wrong_name(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		with self.assertRaisesRegex(TypeError, r'Name must be a non-empty string, not 1'):
			memory_map.add_resource('a', name = 1, size = 0)
		with self.assertRaisesRegex(TypeError, r'Name must be a non-empty string, not \'\''):
			memory_map.add_resource('a', name = '', size = 0)

	def test_add_resource_wrong_name_conflict(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		memory_map.add_resource('a', name = 'foo', size = 0)
		with self.assertRaisesRegex(ValueError, r'Name foo is already used by \'a\''):
			memory_map.add_resource('b', name = 'foo', size = 0)

	def test_add_resource_wrong_address(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Address must be a non-negative integer, not -1'
		):
			memory_map.add_resource('a', name = 'foo', size = 1, addr = -1)

	def test_add_resource_wrong_address_unaligned(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8, alignment = 1)
		with self.assertRaisesRegex(
			ValueError,
			r'Explicitly specified address 0x1 must be a multiple of 0x2 bytes'
		):
			memory_map.add_resource('a', name = 'foo', size = 1, addr = 1)

	def test_add_resource_wrong_size(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Size must be a non-negative integer, not -1'
		):
			memory_map.add_resource('a', name = 'foo', size = -1)

	def test_add_resource_wrong_alignment(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Alignment must be a non-negative integer, not -1'
		):
			memory_map.add_resource('a', name = 'foo', size = 1, alignment = -1)

	def test_add_resource_wrong_out_of_bounds(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x10000\.\.0x10001 out of bounds for memory map spanning '
			r'range 0x0\.\.0x10000 \(16 address bits\)'
		):
			memory_map.add_resource('a', name = 'foo', addr = 0x10000, size = 1)
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x0\.\.0x10001 out of bounds for memory map spanning '
			r'range 0x0\.\.0x10000 \(16 address bits\)'
		):
			memory_map.add_resource('a', name = 'foo', size = 0x10001)

	def test_add_resource_wrong_overlap(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		memory_map.add_resource('a', name = 'foo', size = 16)
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0xa\.\.0xb overlaps with resource \'a\' at 0x0\.\.0x10'
		):
			memory_map.add_resource('b', name = 'bar', size = 1, addr = 10)

	def test_add_resource_wrong_twice(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		memory_map.add_resource('a', name = 'foo', size = 16)
		with self.assertRaisesRegex(
			ValueError,
			r'Resource \'a\' is already added at address range 0x0..0x10'
		):
			memory_map.add_resource('a', name = 'bar', size = 16)

	def test_iter_resources(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		memory_map.add_resource('a', name = 'foo', size = 1)
		memory_map.add_resource('b', name = 'bar', size = 2)
		self.assertEqual(list(memory_map.resources()), [
			('a', 'foo', (0, 1)),
			('b', 'bar', (1, 3)),
		])

	def test_add_window(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 1), (0, 1))
		self.assertEqual(
			memory_map.add_window(MemoryMap(addr_width = 10, data_width = 8)),
			(0x400, 0x800, 1)
		)
		self.assertEqual(memory_map.add_resource('b', name = 'bar', size = 1), (0x800, 0x801))

	def test_add_window_sparse(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 32)
		self.assertEqual(
			memory_map.add_window(
				MemoryMap(addr_width = 10, data_width = 8),
				sparse = True
			),
			(0, 0x400, 1)
		)

	def test_add_window_dense(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 32)
		self.assertEqual(
			memory_map.add_window(
				MemoryMap(addr_width = 10, data_width = 8),
				sparse = False
			),
			(0, 0x100, 4)
		)

	def test_add_window_extend(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(
			memory_map.add_window(
				MemoryMap(addr_width = 17, data_width = 8),
				extend = True
			),
			(0, 0x20000, 1)
		)
		self.assertEqual(memory_map.addr_width, 18)

	def test_add_window_wrong_frozen(self):
		memory_map = MemoryMap(addr_width = 2, data_width = 8)
		memory_map.freeze()
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has been frozen. Cannot add window '
			r'<torii\.lib\.soc\.memory\.MemoryMap object at .+?>'
		):
			memory_map.add_window(MemoryMap(addr_width = 1, data_width = 8))

	def test_add_window_wrong_window(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Window must be a MemoryMap, not \'a\''
		):
			memory_map.add_window(window = 'a')

	def test_add_window_wrong_wider(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Window has data width 16, and cannot be added to a memory map '
			r'with data width 8'
		):
			memory_map.add_window(MemoryMap(addr_width = 10, data_width = 16))

	def test_add_window_wrong_no_mode(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 16)
		with self.assertRaisesRegex(
			ValueError,
			r'Address translation mode must be explicitly specified when adding '
			r'a window with data width 8 to a memory map with data width 16'
		):
			memory_map.add_window(MemoryMap(addr_width = 10, data_width = 8))

	def test_add_window_wrong_ratio(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 16)
		with self.assertRaisesRegex(
			ValueError,
			r'Dense addressing cannot be used because the memory map data width '
			r'16 is not an integer multiple of window data width 7'
		):
			memory_map.add_window(MemoryMap(addr_width = 10, data_width = 7), sparse = False)

	def test_add_window_wrong_out_of_bounds(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x0\.\.0x20000 out of bounds for memory map spanning '
			r'range 0x0\.\.0x10000 \(16 address bits\)'
		):
			memory_map.add_window(MemoryMap(addr_width = 17, data_width = 8))

	def test_add_window_wrong_overlap(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		memory_map.add_window(MemoryMap(addr_width = 10, data_width = 8))
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x200\.\.0x600 overlaps with window '
			r'<torii\.lib\.soc\.memory\.MemoryMap object at .+?> at 0x0\.\.0x400'
		):
			memory_map.add_window(MemoryMap(addr_width = 10, data_width = 8), addr = 0x200)

	def test_add_window_wrong_twice(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		window = MemoryMap(addr_width = 10, data_width = 8)
		memory_map.add_window(window)
		with self.assertRaisesRegex(
			ValueError,
			r'Window <torii\.lib\.soc\.memory\.MemoryMap object at .+?> is already added '
			r'at address range 0x0\.\.0x400'
		):
			memory_map.add_window(window)

	def test_add_window_wrong_name_conflict(self):
		memory_map = MemoryMap(addr_width = 2, data_width = 8)
		memory_map.add_resource('a', name = 'foo', size = 0)
		window = MemoryMap(addr_width = 1, data_width = 8, name = 'foo')
		with self.assertRaisesRegex(ValueError, r'Name foo is already used by \'a\''):
			memory_map.add_window(window)

	def test_add_window_wrong_name_conflict_subordinate(self):
		memory_map = MemoryMap(addr_width = 2, data_width = 8)
		memory_map.add_resource('a', name = 'foo', size = 0)
		memory_map.add_resource('b', name = 'bar', size = 0)
		window = MemoryMap(addr_width = 1, data_width = 8, name = None)
		window.add_resource('c', name = 'foo', size = 0)
		window.add_resource('d', name = 'bar', size = 0)
		with self.assertRaisesRegex(
			ValueError,
			r'The following names are already used: '
			r'bar is used by \'b\'; '
			r'foo is used by \'a\''
		):
			memory_map.add_window(window)

	def test_iter_windows(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 16)
		window_1 = MemoryMap(addr_width = 10, data_width = 8)
		memory_map.add_window(window_1, sparse = False)
		window_2 = MemoryMap(addr_width = 12, data_width = 16)
		memory_map.add_window(window_2)
		self.assertEqual(list(memory_map.windows()), [
			(window_1, (0, 0x200, 2)),
			(window_2, (0x1000, 0x2000, 1)),
		])

	def test_iter_window_patterns(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 16)
		window_1 = MemoryMap(addr_width = 10, data_width = 8)
		memory_map.add_window(window_1, sparse = False)
		window_2 = MemoryMap(addr_width = 12, data_width = 16)
		memory_map.add_window(window_2)
		self.assertEqual(list(memory_map.window_patterns()), [
			(window_1, ('000000----------', 2)),
			(window_2, ('0001------------', 1)),
		])

	def test_iter_window_patterns_covered(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		window = MemoryMap(addr_width = 16, data_width = 8)
		memory_map.add_window(window)
		self.assertEqual(list(memory_map.window_patterns()), [
			(window, ('----------------', 1)),
		])

	def test_align_to(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		self.assertEqual(memory_map.add_resource('a', name = 'foo', size = 1), (0, 1))
		self.assertEqual(memory_map.align_to(10), 0x400)
		self.assertEqual(memory_map.add_resource('b', name = 'bar', size = 16), (0x400, 0x410))

	def test_align_to_wrong(self):
		memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Alignment must be a non-negative integer, not -1'
		):
			memory_map.align_to(alignment = -1)


class MemoryMapDiscoveryTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.root = MemoryMap(addr_width = 32, data_width = 32)
		self.res1 = 'res1'
		self.root.add_resource(self.res1, name = 'name1', size = 16)
		self.win1 = MemoryMap(addr_width = 16, data_width = 32)
		self.res2 = 'res2'
		self.win1.add_resource(self.res2, name = 'name2', size = 32)
		self.res3 = 'res3'
		self.win1.add_resource(self.res3, name = 'name3', size = 32)
		self.root.add_window(self.win1)
		self.res4 = 'res4'
		self.root.add_resource(self.res4, name = 'name4', size = 1)
		self.win2 = MemoryMap(addr_width = 16, data_width = 8)
		self.res5 = 'res5'
		self.win2.add_resource(self.res5, name = 'name5', size = 16)
		self.root.add_window(self.win2, sparse = True)
		self.win3 = MemoryMap(addr_width = 16, data_width = 8, name = 'win3')
		self.res6 = 'res6'
		self.win3.add_resource(self.res6, name = 'name6', size = 16)
		self.root.add_window(self.win3, sparse = False)

	def test_iter_all_resources(self):
		res_info = list(self.root.all_resources())

		self.assertIs(res_info[0].resource, self.res1)
		self.assertEqual(res_info[0].name,  ('name1',))
		self.assertEqual(res_info[0].start, 0x00000000)
		self.assertEqual(res_info[0].end,   0x00000010)
		self.assertEqual(res_info[0].width, 32)

		self.assertIs(res_info[1].resource, self.res2)
		self.assertEqual(res_info[1].name,  ('name2',))
		self.assertEqual(res_info[1].start, 0x00010000)
		self.assertEqual(res_info[1].end,   0x00010020)
		self.assertEqual(res_info[1].width, 32)

		self.assertIs(res_info[2].resource, self.res3)
		self.assertEqual(res_info[2].name,  ('name3',))
		self.assertEqual(res_info[2].start, 0x00010020)
		self.assertEqual(res_info[2].end,   0x00010040)
		self.assertEqual(res_info[2].width, 32)

		self.assertIs(res_info[3].resource, self.res4)
		self.assertEqual(res_info[3].name,  ('name4',))
		self.assertEqual(res_info[3].start, 0x00020000)
		self.assertEqual(res_info[3].end,   0x00020001)
		self.assertEqual(res_info[3].width, 32)

		self.assertIs(res_info[4].resource, self.res5)
		self.assertEqual(res_info[4].name,  ('name5',))
		self.assertEqual(res_info[4].start, 0x00030000)
		self.assertEqual(res_info[4].end,   0x00030010)
		self.assertEqual(res_info[4].width, 8)

		self.assertIs(res_info[5].resource, self.res6)
		self.assertEqual(res_info[5].name,  ('win3', 'name6'))
		self.assertEqual(res_info[5].start, 0x00040000)
		self.assertEqual(res_info[5].end,   0x00040004)
		self.assertEqual(res_info[5].width, 32)

	def test_find_resource(self):
		for res_info in self.root.all_resources():
			other = self.root.find_resource(res_info.resource)
			self.assertIs(other.resource, res_info.resource)
			self.assertEqual(other.name,  res_info.name)
			self.assertEqual(other.start, res_info.start)
			self.assertEqual(other.end,   res_info.end)
			self.assertEqual(other.width, res_info.width)

	def test_find_resource_wrong(self):
		with self.assertRaises(KeyError) as error:
			self.root.find_resource('resNA')
		self.assertEqual(error.exception.args, ('resNA',))

	def test_decode_address(self):
		for res_info in self.root.all_resources():
			self.assertEqual(self.root.decode_address(res_info.start),   res_info.resource)
			self.assertEqual(self.root.decode_address(res_info.end - 1), res_info.resource)

	def test_decode_address_missing(self):
		self.assertIsNone(self.root.decode_address(address = 0x00000100))



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, Instance
from torii.back import verilog


class System(Elaboratable):
	def __init__(self):
		self.adr   = Signal(16)
		self.dat_r = Signal(8)
		self.dat_w = Signal(8)
		self.we    = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		m.submodules.cpu = Instance(
			'CPU',
			p_RESET_ADDR = 0xfff0,
			i_d_adr      = self.adr,
			i_d_dat_r    = self.dat_r,
			o_d_dat_w    = self.dat_w,
			i_d_we       = self.we,
		)
		return m


if __name__ == "__main__":
	sys = System()

	print(verilog.convert(sys, ports = [sys.adr, sys.dat_r, sys.dat_w, sys.we]))



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl           import *
from torii.asserts       import *
from torii.sim           import *
from torii.lib.scheduler import *

from ..utils             import ToriiTestSuiteCase

class RoundRobinTestCase(ToriiTestSuiteCase):
	def test_count(self):
		dut = RoundRobin(count = 32)
		self.assertEqual(dut.count, 32)
		self.assertEqual(len(dut.requests), 32)
		self.assertEqual(len(dut.grant), 5)

	def test_wrong_count(self):
		with self.assertRaisesRegex(
			ValueError, r'Count must be a non-negative integer, not \'foo\''
		):
			RoundRobin(count = 'foo')
		with self.assertRaisesRegex(
			ValueError, r'Count must be a non-negative integer, not -1'
		):
			RoundRobin(count = -1)


class RoundRobinSimulationTestCase(ToriiTestSuiteCase):
	def test_count_one(self):
		dut = RoundRobin(count = 1)
		sim = Simulator(dut)

		def process():
			yield dut.requests.eq(0)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 0)
			self.assertFalse((yield dut.valid))

			yield dut.requests.eq(1)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 0)
			self.assertTrue((yield dut.valid))
		sim.add_sync_process(process)
		sim.add_clock(1e-6)
		with sim.write_vcd('test.vcd'):
			sim.run()

	def test_transitions(self):
		dut = RoundRobin(count = 3)
		sim = Simulator(dut)

		def process():
			yield dut.requests.eq(0b111)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 1)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b110)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 2)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b010)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 1)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b011)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 0)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b001)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 0)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b101)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 2)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b100)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 2)
			self.assertTrue((yield dut.valid))

			yield dut.requests.eq(0b000)
			yield
			yield Delay(1e-8)
			self.assertFalse((yield dut.valid))

			yield dut.requests.eq(0b001)
			yield
			yield Delay(1e-8)
			self.assertEqual((yield dut.grant), 0)
			self.assertTrue((yield dut.valid))
		sim.add_sync_process(process)
		sim.add_clock(1e-6)
		with sim.write_vcd('test.vcd'):
			sim.run()



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl.ast  import *
from torii.hdl.cd   import *
from torii.hdl.ir   import *
from torii.hdl.xfrm import *
from torii.hdl.mem  import *

from ..utils        import ToriiTestSuiteCase


class DomainRenamerTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.s1 = Signal()
		self.s2 = Signal()
		self.s3 = Signal()
		self.s4 = Signal()
		self.s5 = Signal()
		self.c1 = Signal()

	def test_rename_signals(self):
		f = Fragment()
		f.add_statements(
			self.s1.eq(ClockSignal()),
			ResetSignal().eq(self.s2),
			self.s3.eq(0),
			self.s4.eq(ClockSignal('other')),
			self.s5.eq(ResetSignal('other')),
		)
		f.add_driver(self.s1, None)
		f.add_driver(self.s2, None)
		f.add_driver(self.s3, 'sync')

		f = DomainRenamer('pix')(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (clk pix))
			(eq (rst pix) (sig s2))
			(eq (sig s3) (const 1'd0))
			(eq (sig s4) (clk other))
			(eq (sig s5) (rst other))
		)
		''')
		self.assertEqual(f.drivers, {
			None: SignalSet((self.s1, self.s2)),
			'pix': SignalSet((self.s3,)),
		})

	def test_rename_multi(self):
		f = Fragment()
		f.add_statements(
			self.s1.eq(ClockSignal()),
			self.s2.eq(ResetSignal('other')),
		)

		f = DomainRenamer({'sync': 'pix', 'other': 'pix2'})(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (clk pix))
			(eq (sig s2) (rst pix2))
		)
		''')

	def test_rename_cd(self):
		cd_sync = ClockDomain()
		cd_pix  = ClockDomain()

		f = Fragment()
		f.add_domains(cd_sync, cd_pix)

		f = DomainRenamer('ext')(f)
		self.assertEqual(cd_sync.name, 'ext')
		self.assertEqual(f.domains, {
			'ext': cd_sync,
			'pix': cd_pix,
		})

	def test_rename_cd_preserves_allow_reset_less(self):
		cd_pix  = ClockDomain(reset_less = True)

		f = Fragment()
		f.add_domains(cd_pix)
		f.add_statements(
			self.s1.eq(ResetSignal(allow_reset_less = True)),
		)

		f = DomainRenamer('pix')(f)
		f = DomainLowerer()(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (const 1'd0))
		)
		''')


	def test_rename_cd_subfragment(self):
		cd_sync = ClockDomain()
		cd_pix  = ClockDomain()

		f1 = Fragment()
		f1.add_domains(cd_sync, cd_pix)
		f2 = Fragment()
		f2.add_domains(cd_sync)
		f1.add_subfragment(f2)

		f1 = DomainRenamer('ext')(f1)
		self.assertEqual(cd_sync.name, 'ext')
		self.assertEqual(f1.domains, {
			'ext': cd_sync,
			'pix': cd_pix,
		})

	def test_rename_wrong_to_comb(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Domain \'sync\' may not be renamed to \'comb\'$'
		):
			DomainRenamer('comb')

	def test_rename_wrong_from_comb(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Domain \'comb\' may not be renamed$'
		):
			DomainRenamer({'comb': 'sync'})


class DomainLowererTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.s = Signal()

	def test_lower_clk(self):
		sync = ClockDomain()
		f = Fragment()
		f.add_domains(sync)
		f.add_statements(
			self.s.eq(ClockSignal('sync'))
		)

		f = DomainLowerer()(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s) (sig clk))
		)
		''')

	def test_lower_rst(self):
		sync = ClockDomain()
		f = Fragment()
		f.add_domains(sync)
		f.add_statements(
			self.s.eq(ResetSignal('sync'))
		)

		f = DomainLowerer()(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s) (sig rst))
		)
		''')

	def test_lower_rst_reset_less(self):
		sync = ClockDomain(reset_less = True)
		f = Fragment()
		f.add_domains(sync)
		f.add_statements(
			self.s.eq(ResetSignal('sync', allow_reset_less = True))
		)

		f = DomainLowerer()(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s) (const 1'd0))
		)
		''')

	def test_lower_drivers(self):
		sync = ClockDomain()
		pix = ClockDomain()
		f = Fragment()
		f.add_domains(sync, pix)
		f.add_driver(ClockSignal('pix'), None)
		f.add_driver(ResetSignal('pix'), 'sync')

		f = DomainLowerer()(f)
		self.assertEqual(f.drivers, {
			None: SignalSet((pix.clk,)),
			'sync': SignalSet((pix.rst,))
		})

	def test_lower_wrong_domain(self):
		f = Fragment()
		f.add_statements(
			self.s.eq(ClockSignal('xxx'))
		)

		with self.assertRaisesRegex(
			DomainError,
			r'^Signal \(clk xxx\) refers to nonexistent domain \'xxx\'$'
		):
			DomainLowerer()(f)

	def test_lower_wrong_reset_less_domain(self):
		sync = ClockDomain(reset_less = True)
		f = Fragment()
		f.add_domains(sync)
		f.add_statements(
			self.s.eq(ResetSignal('sync'))
		)

		with self.assertRaisesRegex(
			DomainError,
			r'^Signal \(rst sync\) refers to reset of reset-less domain \'sync\'$'
		):
			DomainLowerer()(f)


class SampleLowererTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.i = Signal()
		self.o1 = Signal()
		self.o2 = Signal()
		self.o3 = Signal()

	def test_lower_signal(self):
		f = Fragment()
		f.add_statements(
			self.o1.eq(Sample(self.i, 2, 'sync')),
			self.o2.eq(Sample(self.i, 1, 'sync')),
			self.o3.eq(Sample(self.i, 1, 'pix')),
		)

		f = SampleLowerer()(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig o1) (sig $sample$s$i$sync$2))
			(eq (sig o2) (sig $sample$s$i$sync$1))
			(eq (sig o3) (sig $sample$s$i$pix$1))
			(eq (sig $sample$s$i$sync$1) (sig i))
			(eq (sig $sample$s$i$sync$2) (sig $sample$s$i$sync$1))
			(eq (sig $sample$s$i$pix$1) (sig i))
		)
		''')
		self.assertEqual(len(f.drivers['sync']), 2)
		self.assertEqual(len(f.drivers['pix']), 1)

	def test_lower_const(self):
		f = Fragment()
		f.add_statements(
			self.o1.eq(Sample(1, 2, 'sync')),
		)

		f = SampleLowerer()(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig o1) (sig $sample$c$1$sync$2))
			(eq (sig $sample$c$1$sync$1) (const 1'd1))
			(eq (sig $sample$c$1$sync$2) (sig $sample$c$1$sync$1))
		)
		''')
		self.assertEqual(len(f.drivers['sync']), 2)


class SwitchCleanerTestCase(ToriiTestSuiteCase):
	def test_clean(self):
		a = Signal()
		b = Signal()
		stmts = [
			Switch(a, {
				1: a.eq(0),
				0: [
					b.eq(1),
					Switch(b, {1: [
						Switch(a | b, {})
					]})
				]
			})
		]

		self.assertRepr(SwitchCleaner()(stmts), '''
		(
			(switch (sig a)
				(case 1
					(eq (sig a) (const 1'd0)))
				(case 0
					(eq (sig b) (const 1'd1)))
			)
		)
		''')


class LHSGroupAnalyzerTestCase(ToriiTestSuiteCase):
	def test_no_group_unrelated(self):
		a = Signal()
		b = Signal()
		stmts = [
			a.eq(0),
			b.eq(0),
		]

		groups = LHSGroupAnalyzer()(stmts)
		self.assertEqual(list(groups.values()), [
			SignalSet((a,)),
			SignalSet((b,)),
		])

	def test_group_related(self):
		a = Signal()
		b = Signal()
		stmts = [
			a.eq(0),
			Cat(a, b).eq(0),
		]

		groups = LHSGroupAnalyzer()(stmts)
		self.assertEqual(list(groups.values()), [
			SignalSet((a, b)),
		])

	def test_no_loops(self):
		a = Signal()
		b = Signal()
		stmts = [
			a.eq(0),
			Cat(a, b).eq(0),
			Cat(a, b).eq(0),
		]

		groups = LHSGroupAnalyzer()(stmts)
		self.assertEqual(list(groups.values()), [
			SignalSet((a, b)),
		])

	def test_switch(self):
		a = Signal()
		b = Signal()
		stmts = [
			a.eq(0),
			Switch(a, {
				1: b.eq(0),
			})
		]

		groups = LHSGroupAnalyzer()(stmts)
		self.assertEqual(list(groups.values()), [
			SignalSet((a,)),
			SignalSet((b,)),
		])

	def test_lhs_empty(self):
		stmts = [
			Cat().eq(0)
		]

		groups = LHSGroupAnalyzer()(stmts)
		self.assertEqual(list(groups.values()), [
		])


class LHSGroupFilterTestCase(ToriiTestSuiteCase):
	def test_filter(self):
		a = Signal()
		b = Signal()
		stmts = [
			Switch(a, {
				1: a.eq(0),
				0: [
					b.eq(1),
					Switch(b, {1: []})
				]
			})
		]

		self.assertRepr(LHSGroupFilter(SignalSet((a,)))(stmts), '''
		(
			(switch (sig a)
				(case 1
					(eq (sig a) (const 1'd0)))
				(case 0 )
			)
		)
		''')

	def test_lhs_empty(self):
		stmts = [
			Cat().eq(0)
		]

		self.assertRepr(LHSGroupFilter(SignalSet())(stmts), '()')


class ResetInserterTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.s1 = Signal()
		self.s2 = Signal(reset = 1)
		self.s3 = Signal(reset = 1, reset_less = True)
		self.c1 = Signal()

	def test_reset_default(self):
		f = Fragment()
		f.add_statements(
			self.s1.eq(1)
		)
		f.add_driver(self.s1, 'sync')

		f = ResetInserter(self.c1)(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (const 1'd1))
			(switch (sig c1)
				(case 1 (eq (sig s1) (const 1'd0)))
			)
		)
		''')

	def test_reset_cd(self):
		f = Fragment()
		f.add_statements(
			self.s1.eq(1),
			self.s2.eq(0),
		)
		f.add_domains(ClockDomain('sync'))
		f.add_driver(self.s1, 'sync')
		f.add_driver(self.s2, 'pix')

		f = ResetInserter({'pix': self.c1})(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (const 1'd1))
			(eq (sig s2) (const 1'd0))
			(switch (sig c1)
				(case 1 (eq (sig s2) (const 1'd1)))
			)
		)
		''')

	def test_reset_value(self):
		f = Fragment()
		f.add_statements(
			self.s2.eq(0)
		)
		f.add_driver(self.s2, 'sync')

		f = ResetInserter(self.c1)(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s2) (const 1'd0))
			(switch (sig c1)
				(case 1 (eq (sig s2) (const 1'd1)))
			)
		)
		''')

	def test_reset_less(self):
		f = Fragment()
		f.add_statements(
			self.s3.eq(0)
		)
		f.add_driver(self.s3, 'sync')

		f = ResetInserter(self.c1)(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s3) (const 1'd0))
			(switch (sig c1)
				(case 1 )
			)
		)
		''')


class EnableInserterTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.s1 = Signal()
		self.s2 = Signal()
		self.s3 = Signal()
		self.c1 = Signal()

	def test_enable_default(self):
		f = Fragment()
		f.add_statements(
			self.s1.eq(1)
		)
		f.add_driver(self.s1, 'sync')

		f = EnableInserter(self.c1)(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (const 1'd1))
			(switch (sig c1)
				(case 0 (eq (sig s1) (sig s1)))
			)
		)
		''')

	def test_enable_cd(self):
		f = Fragment()
		f.add_statements(
			self.s1.eq(1),
			self.s2.eq(0),
		)
		f.add_driver(self.s1, 'sync')
		f.add_driver(self.s2, 'pix')

		f = EnableInserter({'pix': self.c1})(f)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (const 1'd1))
			(eq (sig s2) (const 1'd0))
			(switch (sig c1)
				(case 0 (eq (sig s2) (sig s2)))
			)
		)
		''')

	def test_enable_subfragment(self):
		f1 = Fragment()
		f1.add_statements(
			self.s1.eq(1)
		)
		f1.add_driver(self.s1, 'sync')

		f2 = Fragment()
		f2.add_statements(
			self.s2.eq(1)
		)
		f2.add_driver(self.s2, 'sync')
		f1.add_subfragment(f2)

		f1 = EnableInserter(self.c1)(f1)
		(f2, _), = f1.subfragments
		self.assertRepr(f1.statements, '''
		(
			(eq (sig s1) (const 1'd1))
			(switch (sig c1)
				(case 0 (eq (sig s1) (sig s1)))
			)
		)
		''')
		self.assertRepr(f2.statements, '''
		(
			(eq (sig s2) (const 1'd1))
			(switch (sig c1)
				(case 0 (eq (sig s2) (sig s2)))
			)
		)
		''')

	def test_enable_read_port(self):
		mem = Memory(width = 8, depth = 4)
		mem.read_port(transparent = False)
		f = EnableInserter(self.c1)(mem).elaborate(platform = None)
		self.assertRepr(f.named_ports['RD_EN'][0], '''
		(cat (m (sig c1) (sig mem_r_en) (const 1'd0)))
		''')

	def test_enable_write_port(self):
		mem = Memory(width = 8, depth = 4)
		mem.write_port()
		f = EnableInserter(self.c1)(mem).elaborate(platform = None)
		self.assertRepr(f.named_ports['WR_EN'][0], '''
		(cat (m
			(sig c1)
			(cat
				(cat
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
					(slice (sig mem_w_en) 0:1)
				)
			)
			(const 8'd0)
		))
		''')


class _MockElaboratable(Elaboratable):
	def __init__(self):
		self.s1 = Signal()

	def elaborate(self, platform):
		f = Fragment()
		f.add_statements(
			self.s1.eq(1)
		)
		f.add_driver(self.s1, 'sync')
		return f


class TransformedElaboratableTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.c1 = Signal()
		self.c2 = Signal()

	def test_getattr(self):
		e = _MockElaboratable()
		te = EnableInserter(self.c1)(e)

		self.assertIs(te.s1, e.s1)

	def test_composition(self):
		e = _MockElaboratable()
		te1 = EnableInserter(self.c1)(e)
		te2 = ResetInserter(self.c2)(te1)

		self.assertIsInstance(te1, TransformedElaboratable)
		self.assertIs(te1, te2)

		f = Fragment.get(te2, None)
		self.assertRepr(f.statements, '''
		(
			(eq (sig s1) (const 1'd1))
			(switch (sig c1)
				(case 0 (eq (sig s1) (sig s1)))
			)
			(switch (sig c2)
				(case 1 (eq (sig s1) (const 1'd0)))
			)
		)
		''')



# SPDX-License-Identifier: BSD-2-Clause

from .dsl  import (
	Attrs, Clock, Connector, DiffPairs, DiffPairsN, Pins, PinsN,
	Resource, Subsignal
)
from .plat import Platform, TemplatedPlatform
from .res  import ResourceError

__all__ = (
	'Attrs',
	'Clock',
	'Connector',
	'DiffPairs',
	'DiffPairsN',
	'Pins',
	'PinsN',
	'Platform',
	'Resource',
	'ResourceError',
	'Subsignal',
	'TemplatedPlatform',
)



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal
from torii.sim  import Settle
from torii.test import ToriiTestCase

class TestSyncDUT(Elaboratable):
	def __init__(self) -> None:
		self.a = Signal(reset = 1)
		self.b = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()

		m.d.sync += [
			self.b.eq(self.a),
			self.a.eq(~self.a),
		]

		return m

class TestMultiSyncDUT(Elaboratable):
	def __init__(self) -> None:
		self.a = Signal(reset = 1)
		self.b = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()

		m.d.sync1 += [
			self.a.eq(~self.a),
		]

		m.d.sync2 += [
			self.b.eq(~self.b),
		]

		return m


class ToriiTestCaseSyncTest(ToriiTestCase):
	domains = (('sync', 1e6), )
	dut: TestSyncDUT = TestSyncDUT

	@ToriiTestCase.simulation
	@ToriiTestCase.sync_domain(domain = 'sync')
	def test_sync(self):
		self.assertEqual((yield self.dut.a), 1)
		yield Settle()
		yield
		self.assertEqual((yield self.dut.a), 0)
		self.assertEqual((yield self.dut.b), 1)
		yield Settle()
		yield
		yield Settle()
		yield

class ToriiTestCaseMultiSyncTest(ToriiTestCase):
	domains = (('sync1', 1e6), ('sync2', 2e6), )
	dut: TestMultiSyncDUT = TestMultiSyncDUT

	@ToriiTestCase.simulation
	def test_multi_sync(self):
		@ToriiTestCase.sync_domain(domain = 'sync1')
		def sync1(self):
			self.assertEqual((yield self.dut.a), 1)
			yield Settle()
			yield
			yield from self.step(100)

		@ToriiTestCase.sync_domain(domain = 'sync2')
		def sync2(self):
			self.assertEqual((yield self.dut.b), 0)
			yield Settle()
			yield
			yield from self.step(100)

		# Invoke sync process setup, this won't actually run the sim it just
		# ensures that the process defined by the function is added to the
		# simulator.
		sync1(self)
		sync2(self)



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl                import *
from torii.asserts            import *
from torii.sim                import *
from torii.lib.coding.one_hot import Encoder, Decoder

from ...utils              import ToriiTestSuiteCase

class EncoderTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		enc = Encoder(4)

		def process():
			self.assertEqual((yield enc.n), 1)
			self.assertEqual((yield enc.o), 0)

			yield enc.i.eq(0b0001)
			yield Settle()
			self.assertEqual((yield enc.n), 0)
			self.assertEqual((yield enc.o), 0)

			yield enc.i.eq(0b0100)
			yield Settle()
			self.assertEqual((yield enc.n), 0)
			self.assertEqual((yield enc.o), 2)

			yield enc.i.eq(0b0110)
			yield Settle()
			self.assertEqual((yield enc.n), 1)
			self.assertEqual((yield enc.o), 0)

		sim = Simulator(enc)
		sim.add_process(process)
		sim.run()

class DecoderTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		dec = Decoder(4)

		def process():
			self.assertEqual((yield dec.o), 0b0001)

			yield dec.i.eq(1)
			yield Settle()
			self.assertEqual((yield dec.o), 0b0010)

			yield dec.i.eq(3)
			yield Settle()
			self.assertEqual((yield dec.o), 0b1000)

			yield dec.n.eq(1)
			yield Settle()
			self.assertEqual((yield dec.o), 0b0000)

		sim = Simulator(dec)
		sim.add_process(process)
		sim.run()



# SPDX-License-Identifier: BSD-2-Clause

from .ast  import (
	Array, Cat, ClockSignal, Const, Mux, Repl, ResetSignal,
	Shape, Signal, Value, signed, unsigned
)
from .cd   import ClockDomain
from .dsl  import Module
from .ir   import Elaboratable, Fragment, Instance
from .mem  import Memory
from .rec  import Record
from .xfrm import DomainRenamer, EnableInserter, ResetInserter

__all__ = (
	'Array',
	'Cat',
	'ClockDomain',
	'ClockSignal',
	'Const',
	'DomainRenamer',
	'Elaboratable',
	'EnableInserter',
	'Fragment',
	'Instance',
	'Memory',
	'Module',
	'Mux',
	'Record',
	'Repl',
	'ResetInserter',
	'ResetSignal',
	'Shape',
	'Signal',
	'signed',
	'unsigned',
	'Value',
)



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, Cat
from torii.back import verilog


class UARTReceiver(Elaboratable):
	def __init__(self, divisor: int):
		self.divisor = divisor

		self.i    = Signal()
		self.data = Signal(8)
		self.rdy  = Signal()
		self.ack  = Signal()
		self.err  = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()

		ctr = Signal(range(self.divisor))
		stb = Signal()
		with m.If(ctr == 0):
			m.d.sync += ctr.eq(self.divisor - 1)
			m.d.comb += stb.eq(1)
		with m.Else():
			m.d.sync += ctr.eq(ctr - 1)

		bit = Signal(3)
		with m.FSM() as fsm:
			with m.State('START'):
				with m.If(~self.i):
					m.next = 'DATA'
					m.d.sync += [
						ctr.eq(self.divisor // 2),
						bit.eq(7),
					]
			with m.State('DATA'):
				with m.If(stb):
					m.d.sync += [
						bit.eq(bit - 1),
						self.data.eq(Cat(self.i, self.data))
					]
					with m.If(bit == 0):
						m.next = 'STOP'
			with m.State('STOP'):
				with m.If(stb):
					with m.If(self.i):
						m.next = 'DONE'
					with m.Else():
						m.next = 'ERROR'

			with m.State('DONE'):
				m.d.comb += self.rdy.eq(1)
				with m.If(self.ack):
					m.next = 'START'

			m.d.comb += self.err.eq(fsm.ongoing('ERROR'))
			with m.State('ERROR'):
				pass

		return m


if __name__ == '__main__':
	rx = UARTReceiver(20)

	print(verilog.convert(rx, ports = [rx.i, rx.data, rx.rdy, rx.ack, rx.err]))



# SPDX-License-Identifier: BSD-2-Clause

from contextlib import contextmanager
from itertools  import chain
from re         import search

from vcd        import VCDWriter
from vcd.gtkw   import GTKWSave

from ..hdl      import ClockSignal, ResetSignal
from ..hdl.ast  import SignalDict
from ._base     import BaseEngine, BaseSignalState, BaseSimulation
from ._pyclock  import PyClockProcess
from ._pycoro   import PyCoroProcess
from ._pyrtl    import _FragmentCompiler

__all__ = (
	'PySimEngine',
)


class _NameExtractor:
	def __init__(self):
		self.names = SignalDict()

	def __call__(self, fragment, *, hierarchy = ('bench', 'top',)):
		def add_signal_name(signal):
			hierarchical_signal_name = (*hierarchy, signal.name)
			if signal not in self.names:
				self.names[signal] = {hierarchical_signal_name}
			else:
				self.names[signal].add(hierarchical_signal_name)

		for domain_name, domain_signals in fragment.drivers.items():
			if domain_name is not None:
				domain = fragment.domains[domain_name]
				add_signal_name(domain.clk)
				if domain.rst is not None:
					add_signal_name(domain.rst)

		for statement in fragment.statements:
			for signal in statement._lhs_signals() | statement._rhs_signals():
				if not isinstance(signal, (ClockSignal, ResetSignal)):
					add_signal_name(signal)

		for subfragment_index, (subfragment, subfragment_name) in enumerate(fragment.subfragments):
			if subfragment_name is None:
				subfragment_name = f'U${subfragment_name}'
			self(subfragment, hierarchy = (*hierarchy, subfragment_name))

		return self.names


class _VCDWriter:
	@staticmethod
	def decode_to_vcd(signal, value):
		return signal.decoder(value).expandtabs().replace(' ', '_')

	def __init__(self, fragment, *, vcd_file, gtkw_file = None, traces = ()):
		if isinstance(vcd_file, str):
			vcd_file = open(vcd_file, 'wt')
		if isinstance(gtkw_file, str):
			gtkw_file = open(gtkw_file, 'wt')

		self.vcd_vars = SignalDict()
		self.vcd_file = vcd_file
		self.vcd_writer = vcd_file and VCDWriter(self.vcd_file,
			timescale = '1 ps', comment = 'Generated by Torii')

		self.gtkw_names = SignalDict()
		self.gtkw_file = gtkw_file
		self.gtkw_save = gtkw_file and GTKWSave(self.gtkw_file)

		self.traces = []

		signal_names = _NameExtractor()(fragment)

		trace_names = SignalDict()
		for trace in traces:
			if trace not in signal_names:
				trace_names[trace] = { ('bench', trace.name) }
			self.traces.append(trace)

		if self.vcd_writer is None:
			return

		for signal, names in chain(signal_names.items(), trace_names.items()):
			if signal.decoder:
				var_type = 'string'
				var_size = 1
				var_init = self.decode_to_vcd(signal, signal.reset)
			else:
				var_type = 'wire'
				var_size = signal.width
				var_init = signal.reset

			for (*var_scope, var_name) in names:
				if search(r'[ \t\r\n]', var_name):
					raise NameError(f'Signal \'{".".join(var_scope)}.{var_name}\' contains a whitespace character')

				suffix = None
				while True:
					try:
						if suffix is None:
							var_name_suffix = var_name
						else:
							var_name_suffix = f'{var_name}${suffix}'
						if signal not in self.vcd_vars:
							vcd_var = self.vcd_writer.register_var(
								scope = var_scope, name = var_name_suffix,
								var_type = var_type, size = var_size, init = var_init)
							self.vcd_vars[signal] = vcd_var
						else:
							self.vcd_writer.register_alias(
								scope = var_scope, name = var_name_suffix,
								var = self.vcd_vars[signal])
						break
					except KeyError:
						suffix = (suffix or 0) + 1

				if signal not in self.gtkw_names:
					self.gtkw_names[signal] = (*var_scope, var_name_suffix)

	def update(self, timestamp, signal, value):
		vcd_var = self.vcd_vars.get(signal)
		if vcd_var is None:
			return

		if signal.decoder:
			var_value = self.decode_to_vcd(signal, value)
		else:
			var_value = value
		self.vcd_writer.change(vcd_var, timestamp, var_value)

	def close(self, timestamp):
		if self.vcd_writer is not None:
			self.vcd_writer.close(timestamp)

		if self.gtkw_save is not None:
			self.gtkw_save.dumpfile(self.vcd_file.name)
			self.gtkw_save.dumpfile_size(self.vcd_file.tell())

			self.gtkw_save.treeopen('top')
			for signal in self.traces:
				if len(signal) > 1 and not signal.decoder:
					suffix = f'[{len(signal) - 1}:0]'
				else:
					suffix = ''
				self.gtkw_save.trace('.'.join(self.gtkw_names[signal]) + suffix)

		if self.vcd_file is not None:
			self.vcd_file.close()
		if self.gtkw_file is not None:
			self.gtkw_file.close()


class _Timeline:
	def __init__(self):
		self.now = 0
		self.deadlines = dict()

	def reset(self):
		self.now = 0
		self.deadlines.clear()

	def at(self, run_at, process):
		assert process not in self.deadlines
		self.deadlines[process] = run_at

	def delay(self, delay_by, process):
		if delay_by is None:
			run_at = self.now
		else:
			run_at = self.now + delay_by
		self.at(run_at, process)

	def advance(self):
		nearest_processes = set()
		nearest_deadline = None
		for process, deadline in self.deadlines.items():
			if deadline is None:
				if nearest_deadline is not None:
					nearest_processes.clear()
				nearest_processes.add(process)
				nearest_deadline = self.now
				break
			elif nearest_deadline is None or deadline <= nearest_deadline:
				assert deadline >= self.now
				if nearest_deadline is not None and deadline < nearest_deadline:
					nearest_processes.clear()
				nearest_processes.add(process)
				nearest_deadline = deadline

		if not nearest_processes:
			return False

		for process in nearest_processes:
			process.runnable = True
			del self.deadlines[process]
		self.now = nearest_deadline

		return True


class _PySignalState(BaseSignalState):
	__slots__ = ('signal', 'curr', 'next', 'waiters', 'pending')

	def __init__(self, signal, pending):
		self.signal = signal
		self.pending = pending
		self.waiters = dict()
		self.curr = self.next = signal.reset

	def set(self, value):
		if self.next == value:
			return
		self.next = value
		self.pending.add(self)

	def commit(self):
		if self.curr == self.next:
			return False
		self.curr = self.next

		awoken_any = False
		for process, trigger in self.waiters.items():
			if trigger is None or trigger == self.curr:
				process.runnable = awoken_any = True
		return awoken_any


class _PySimulation(BaseSimulation):
	def __init__(self):
		self.timeline = _Timeline()
		self.signals  = SignalDict()
		self.slots    = []
		self.pending  = set()

	def reset(self):
		self.timeline.reset()
		for signal, index in self.signals.items():
			self.slots[index].curr = self.slots[index].next = signal.reset
		self.pending.clear()

	def get_signal(self, signal):
		try:
			return self.signals[signal]
		except KeyError:
			index = len(self.slots)
			self.slots.append(_PySignalState(signal, self.pending))
			self.signals[signal] = index
			return index

	def add_trigger(self, process, signal, *, trigger = None):
		index = self.get_signal(signal)
		assert (process not in self.slots[index].waiters or
				self.slots[index].waiters[process] == trigger)
		self.slots[index].waiters[process] = trigger

	def remove_trigger(self, process, signal):
		index = self.get_signal(signal)
		assert process in self.slots[index].waiters
		del self.slots[index].waiters[process]

	def wait_interval(self, process, interval):
		self.timeline.delay(interval, process)

	def commit(self, changed = None):
		converged = True
		for signal_state in self.pending:
			if signal_state.commit():
				converged = False
		if changed is not None:
			changed.update(self.pending)
		self.pending.clear()
		return converged


class PySimEngine(BaseEngine):
	def __init__(self, fragment):
		self._state = _PySimulation()
		self._timeline = self._state.timeline

		self._fragment = fragment
		self._processes = _FragmentCompiler(self._state)(self._fragment)
		self._vcd_writers = []

	def add_coroutine_process(self, process, *, default_cmd):
		self._processes.add(
			PyCoroProcess(self._state, self._fragment.domains, process, default_cmd = default_cmd)
		)

	def add_clock_process(self, clock, *, phase, period):
		self._processes.add(
			PyClockProcess(self._state, clock, phase = phase, period = period)
		)

	def reset(self):
		self._state.reset()
		for process in self._processes:
			process.reset()

	def _step(self):
		changed = set() if self._vcd_writers else None

		# Performs the two phases of a delta cycle in a loop:
		converged = False
		while not converged:
			# 1. eval: run and suspend every non-waiting process once, queueing signal changes
			for process in self._processes:
				if process.runnable:
					process.runnable = False
					process.run()

			# 2. commit: apply every queued signal change, waking up any waiting processes
			converged = self._state.commit(changed)

		for vcd_writer in self._vcd_writers:
			for signal_state in changed:
				vcd_writer.update(self._timeline.now,
					signal_state.signal, signal_state.curr)

	def advance(self):
		self._step()
		self._timeline.advance()
		return any(not process.passive for process in self._processes)

	@property
	def now(self):
		return self._timeline.now

	@contextmanager
	def write_vcd(self, *, vcd_file, gtkw_file, traces):
		vcd_writer = _VCDWriter(self._fragment,
			vcd_file = vcd_file, gtkw_file = gtkw_file, traces = traces)
		try:
			self._vcd_writers.append(vcd_writer)
			yield
		finally:
			vcd_writer.close(self._timeline.now)
			self._vcd_writers.remove(vcd_writer)



# SPDX-License-Identifier: BSD-2-Clause

from .ecp5        import ECP5Platform        as LatticeECP5Platform
from .ice40       import ICE40Platform       as LatticeICE40Platform
from .machxo_2_3l import MachXO2Or3LPlatform as LatticeMachXO2Or3LPlatform
from .machxo_2_3l import MachXO2Platform     as LatticeMachXO2Platform
from .machxo_2_3l import MachXO3LPlatform    as LatticeMachXO3LPlatform


__all__ = (
	'LatticeECP5Platform',
	'LatticeICE40Platform',
	'LatticeMachXO2Or3LPlatform',
	'LatticeMachXO2Platform',
	'LatticeMachXO3LPlatform',
)



# SPDX-License-Identifier: BSD-2-Clause

from collections     import OrderedDict
from collections.abc import Mapping
from typing          import Generator, Optional

from ...util.units   import bits_for
from .               import event
from .memory         import MemoryMap

__all__ = (
	'ConstantBool',
	'ConstantInt',
	'ConstantMap',
	'ConstantValue',
	'PeripheralInfo',
)


class ConstantValue:
	pass


class ConstantBool(ConstantValue):
	'''
	Boolean constant.

	Parameters
	----------
	value : bool
		Constant value.

	'''

	def __init__(self, value: bool) -> None:
		if not isinstance(value, bool):
			raise TypeError(f'Value must be a bool, not {value!r}')
		self._value = value

	@property
	def value(self) -> bool:
		return self._value

	def __repr__(self) -> str:
		return f'ConstantBool({self.value})'


class ConstantInt(ConstantValue):
	'''
	Integer constant.

	Parameters
	----------
	value : int
		Constant value.
	width : int
		Width in bits. Optional. ``bits_for(value)`` by default.
	signed : bool
		Signedness. Optional. ``value < 0`` by default.

	'''

	def __init__(
		self, value: int, *, width: Optional[int] = None, signed: Optional[bool] = None
	) -> None:
		if not isinstance(value, int):
			raise TypeError(f'Value must be an integer, not {value!r}')
		self._value = value

		if width is None:
			width = bits_for(value)
		if not isinstance(width, int):
			raise TypeError(f'Width must be an integer, not {width!r}')
		if width < bits_for(value):
			raise ValueError(f'Width must be greater than or equal to the number of bits needed to represent {value}')
		self._width = width

		if signed is None:
			signed = value < 0
		if not isinstance(signed, bool):
			raise TypeError(f'Signedness must be a bool, not {signed!r}')
		self._signed = signed

	@property
	def value(self) -> int:
		return self._value

	@property
	def width(self) -> int:
		return self._width

	@property
	def signed(self) -> bool:
		return self._signed

	def __repr__(self) -> str:
		return f'ConstantInt({self.value}, width={self.width}, signed={self.signed})'


class ConstantMap(Mapping):
	'''
	Named constant map.

	A read-only container for named constants. Keys are iterated in insertion order.

	Parameters
	----------
	**constants : dict(str : :class:`ConstantValue`)
		Named constants.

	Examples
	--------
	>>> ConstantMap(RX_FIFO_DEPTH = 16)
	ConstantMap([('RX_FIFO_DEPTH', ConstantInt(16, width = 5, signed = False))])

	'''

	def __init__(self, **constants: dict[str, ConstantValue]) -> None:
		self._storage = OrderedDict()
		for key, value in constants.items():
			if isinstance(value, bool):
				value = ConstantBool(value)
			if isinstance(value, int):
				value = ConstantInt(value)
			if not isinstance(value, ConstantValue):
				raise TypeError(f'Constant value must be an instance of ConstantValue, not {value!r}')
			self._storage[key] = value

	def __getitem__(self, key) -> ConstantValue:
		return self._storage[key]

	def __iter__(self) -> Generator[ConstantValue, None, None]:
		yield from self._storage

	def __len__(self) -> int:
		return len(self._storage)

	def __repr__(self) -> str:
		return f'ConstantMap({list(self._storage.items())})'


class PeripheralInfo:
	'''
	Peripheral metadata.

	A unified description of the local resources of a peripheral. It may be queried in order to
	recover its memory windows, CSR registers, event sources and configuration constants.

	Parameters
	----------
	memory_map : :class:`MemoryMap`
		Memory map of the peripheral.
	irq : :class:`event.Source`
		IRQ line of the peripheral. Optional.
	constant_map : :class:`ConstantMap`
		Constant map of the peripheral. Optional.

	'''

	def __init__(
		self, *, memory_map: MemoryMap, irq: Optional[event.Source] = None,
		constant_map: Optional[ConstantMap] = None
	) -> None:
		if not isinstance(memory_map, MemoryMap):
			raise TypeError(f'Memory map must be an instance of MemoryMap, not {memory_map!r}')
		memory_map.freeze()
		self._memory_map = memory_map

		if irq is not None and not isinstance(irq, event.Source):
			raise TypeError(f'IRQ line must be an instance of event.Source, not {irq!r}')
		self._irq = irq

		if constant_map is None:
			constant_map = ConstantMap()
		if not isinstance(constant_map, ConstantMap):
			raise TypeError(f'Constant map must be an instance of ConstantMap, not {constant_map!r}')
		self._constant_map = constant_map

	@property
	def memory_map(self) -> MemoryMap:
		'''
		Memory map.

		Return value
		------------
		A :class:`MemoryMap` describing the local address space of the peripheral.

		'''

		return self._memory_map

	@property
	def irq(self) -> event.Source:
		'''
		IRQ line.

		Return value
		------------
		An :class:`event.Source` used by the peripheral to request interrupts. If provided, its
		event map describes local events.

		Exceptions
		----------
		Raises :exn:`NotImplementedError` if the peripheral info does not have an IRQ line.

		'''

		if self._irq is None:
			raise NotImplementedError('Peripheral info does not have an IRQ line')
		return self._irq

	@property
	def constant_map(self) -> ConstantMap:
		'''
		Constant map.

		Return value
		------------
		A :class:`ConstantMap` containing configuration constants of the peripheral.

		'''

		return self._constant_map



# SPDX-License-Identifier: BSD-2-Clause

import logging as log
from functools import wraps
from math      import ceil
from os        import getenv
from pathlib   import Path
from typing    import Optional
from unittest  import TestCase

from ..hdl.ast import Signal
from ..hdl.ir  import Fragment
from ..sim     import Settle, Simulator

__all__ = (
	'ToriiTestCase',
)

class ToriiTestCase(TestCase):
	'''
	Torii test case wrapper for pythons unittest library

	This class wraps the :py:class:`TestCase` class from the `unittest` module
	from the python standard lib. It has useful methods for testing and simulating
	Torii based gateware.

	Attributes
	----------
	domains : tuple[tuple[str, float], ...]
		The collection of clock domains and frequencies
	out_dir : str
		The test output directory.
	dut : Elaboratable
		The elaboratable to test.
	dut_args : dict[str, Any]
		The initialization arguments for the elaboratable.
	platform : MockPlatform, Any
		The platform passed to the Elaboratable DUT

	'''

	domains   = (('sync', 1e8),)
	out_dir   = None
	dut       = None
	dut_args  = {}
	platform  = None

	def __init__(self, *args, **kwargs) -> None:
		super().__init__(*args, **kwargs)

		self._frag = None

	@property
	def vcd_name(self) -> str:
		''' Return the name used to generate VCD files '''
		return f'test-{self.__class__.__name__}'

	def clk_period(self, domain: Optional[str] = None) -> float:
		''' Returns the period of the clock on the given domain '''
		if domain is None:
			return 1 / self.domains[0][1]

		freq = next(d[1] for d in self.domains if d[0] == domain)

		return 1 / freq

	def run_sim(self, *, suffix: Optional[str] = None) -> None:
		'''
		Run the simulation

		If the environment variable ``TORII_TEST_INHIBIT_VCD``
		is set, then no VCDs will be generated.

		Keyword Args
		------------
		suffix : str
			The option VCD test case suffix.

		'''

		if getenv('TORII_TEST_INHIBIT_VCD', default = False):
			self.sim.reset()
			self.sim.run()
		else:
			out_name = f'{self.vcd_name}{f"-{suffix}" if suffix is not None else ""}.vcd'
			# HACK: Converting the path object to a string because write_vcd is dumb
			out_file = str((self.out_dir / out_name).resolve())

			with self.sim.write_vcd(out_file):
				self.sim.reset()
				self.sim.run()

	def setUp(self) -> None:
		''' Set up the simulator per test-case '''
		if self.dut is not None:
			self.dut   = self.init_dut()
			self._frag = Fragment.get(self.dut, self.platform)
			self.sim   = Simulator(self._frag)

			if self.out_dir is None:
				if (Path.cwd() / 'build').exists():
					self.out_dir = Path.cwd() / 'build' / 'tests'
				else:
					self.out_dir = Path.cwd() / 'test-vcds'

			self.out_dir.mkdir(exist_ok = True, parents = True)

			for d, _ in self.domains:
				self.sim.add_clock(self.clk_period(d), domain = d)

	def init_dut(self):
		''' Initialize and return the DUT '''

		return self.dut(**self.dut_args)

	def init_signals(self):

		yield Signal()

	def wait_for(self, time: float, domain: Optional[str] = None):
		'''
		Waits for the number time units.

		Parameters
		----------
		time : float
			The unit of time to wait.

		'''

		c = ceil(time / self.clk_period(domain))

		yield from self.step(c)

	@staticmethod
	def pulse(sig: Signal , *, neg: bool = False, post_step: bool = True):
		'''
		Pulse a given signal.

		Pulse a given signal to 1 then 0, or if `neg` is set to `True` pulse
		from 1 to 0.

		Parameters
		----------
		sig : Signal
			The signal to pulse.

		Keyword Args
		------------
		neg : bool
			Inverts the pulse direction.
		post_step : bool
			Insert additional simulation step after pulse.

		'''

		if not neg:
			yield sig.eq(1)
			yield
			yield sig.eq(0)
		else:
			yield sig.eq(0)
			yield
			yield sig.eq(1)
		if post_step:
			yield


	@staticmethod
	def pulse_pos(sig: Signal, *, post_step: bool = True):
		'''
		Inserts a positive pulse on the given signal

		Parameters
		----------
		sig : Signal
			The signal to pulse.

		Keyword Args
		------------
		post_step : bool
			Insert additional simulation step after pulse.

		'''

		yield from ToriiTestCase.pulse(sig, neg = False, post_step = post_step)

	@staticmethod
	def pulse_neg(sig: Signal, *, post_step: bool = True):
		'''
		Inserts a negative pulse on the given signal

		Parameters
		----------
		sig : Signal
			The signal to pulse.

		Keyword Args
		------------
		post_step : bool
			Insert additional simulation step after pulse.

		'''

		yield from ToriiTestCase.pulse(sig, neg = True, post_step = post_step)

	@staticmethod
	def step(cycles: int):
		'''
		Step simulator.

		This advances the simulation by the given number of cycles.

		Parameters
		----------
		cycles : int
			Number of cycles to step the simulator.

		'''

		for _ in range(cycles + 1):
			yield

	@staticmethod
	def settle(count: int = 1):
		'''
		Settle simulator.

		This advances the simulation by the given number of settles.

		Parameters
		----------
		count : int
			Number of settles to invoke in the simulator.

		'''
		for _ in range(count):
			yield Settle()
			yield

	@staticmethod
	def wait_until_high(strobe: Signal, *, timeout: int = None):
		'''
		Run simulation until signal goes high.

		Runs the simulation while checking for the positive edge of the `strobe`
		signal.

		Will run until a positive edge is seen, or if `timeout` is set, will run
		for at most that many cycles.

		This is the positive counterpart for the :py:func:`wait_until_low` method.

		Parameters
		----------
		strobe : Signal
			The signal to check the strobe for.

		Keyword Args
		------------
		timeout : int
			The max number of cycles to wait.
		'''

		elapsed_cycles = 0
		while not (yield strobe):
			yield
			elapsed_cycles += 1
			if timeout and elapsed_cycles > timeout:
				raise RuntimeError(f'Timeout waiting for \'{strobe.name}\' to go high')

	@staticmethod
	def wait_until_low(strobe: Signal, *, timeout: int = None):
		'''
		Run simulation until signal goes low.

		Runs the simulation while checking for the negative edge of the `strobe`
		signal.

		Will run until a negative edge is seen, or if `timeout` is set, will run
		for at most that many cycles.

		This is the negative counterpart for the :py:func:`wait_until_high` method.

		Parameters
		----------
		strobe : Signal
			The signal to check the strobe for.

		Keyword Args
		------------
		timeout : int
			The max number of cycles to wait.

		'''

		elapsed_cycles = 0
		while (yield strobe):
			yield
			elapsed_cycles += 1
			if timeout and elapsed_cycles > timeout:
				raise RuntimeError(f'Timeout waiting for \'{strobe.name}\' to go low')

	@staticmethod
	def simulation(func):
		'''
		Simulation test case decorator

		..note::
			This must always be the top-most decorator due to how
			python decorator precedence works.

		Parameters
		----------
		func
			The decorated function.

		'''

		def _run(self: ToriiTestCase):
			# Invoke the wrapped setup we did
			func(self)

			log.debug(f'Running simulation for \'{func.__name__}\'')
			self.run_sim(suffix = func.__name__)
		return _run

	@staticmethod
	def sync_domain(*, domain: str):
		'''
		This decorator is used to annotate that the following function should be simulated
		in the specified synchronous domain.

		It should be used in combination with :py:func:`ToriiTestCase.sim_test` to simulate
		a synchronous process.

		Parameters
		----------
		domain : str
			The domain this process belongs to

		'''

		def _sync(func):
			@wraps(func)
			def _add(self: ToriiTestCase):
				if domain not in map(lambda d: d[0], self.domains):
					raise ValueError(
						f'The requested domain \'{domain}\' was not specified in the ',
						'test suite.\n'
						'The following domains are defined:\n',
						f'{self.domains}'
					)


				def proc():
					yield from func(self)

				log.debug(f'Adding synchronous process \'{func.__name__}\' to domain \'{domain}\'')
				self.sim.add_sync_process(proc, domain = domain)

			return _add
		return _sync

	@staticmethod
	def comb_domain(func):
		'''
		This decorator is used to annotate that the following function should be simulated
		in the combinatorial domain.

		It should be used in combination with :py:func:`ToriiTestCase.sim_test` to simulate
		a combinatorial process.
		'''

		@wraps(func)
		def _add_comb(self: ToriiTestCase):

			def proc():
				yield from func(self)

			log.debug(f'Adding combinatorial process \'{func.__name__}\'')
			self.sim.add_process(proc)

		return _add_comb



# SPDX-License-Identifier: BSD-2-Clause

from typing        import Optional

from ..hdl.ast     import SignalDict
from ..tools.yosys import YosysError, find_yosys
from .             import rtlil

__all__ = (
	'convert_fragment',
	'convert',
	'YosysError',
)

def _convert_rtlil_text(
	rtlil_text: str, black_boxes: Optional[dict[str, str]], *, src_loc_at: int = 0
) -> str:
	if black_boxes is not None:
		if not isinstance(black_boxes, dict):
			raise TypeError(f'CXXRTL black boxes must be a dictionary, not {black_boxes!r}')
		for box_name, box_source in black_boxes.items():
			if not isinstance(box_name, str):
				raise TypeError(f'CXXRTL black box name must be a string, not {box_name!r}')
			if not isinstance(box_source, str):
				raise TypeError(f'CXXRTL black box source code must be a string, not {box_source!r}')

	yosys = find_yosys()

	script = []
	if black_boxes is not None:
		for box_name, box_source in black_boxes.items():
			script.append(f'read_rtlil <<rtlil\n{box_source}\nrtlil')
	script.append(f'read_rtlil <<rtlil\n{rtlil_text}\nrtlil')
	script.append('write_cxxrtl')

	return yosys.run(['-q', '-'], '\n'.join(script), src_loc_at = 1 + src_loc_at)


def convert_fragment(*args, black_boxes: Optional[dict[str, str]] = None, **kwargs) -> tuple[str, SignalDict]:
	rtlil_text, name_map = rtlil.convert_fragment(*args, **kwargs)
	return (_convert_rtlil_text(rtlil_text, black_boxes, src_loc_at = 1), name_map)


def convert(*args, black_boxes: Optional[dict[str, str]] = None, **kwargs) -> str:
	rtlil_text = rtlil.convert(*args, **kwargs)
	return _convert_rtlil_text(rtlil_text, black_boxes, src_loc_at = 1)



# SPDX-License-Identifier: BSD-2-Clause

from .display    import Display7SegResource, VGADACResource, VGAResource
from .extensions import (
	PmodDualHBridgeType6Resource, PmodGPIOType1Resource,
	PmodHBridgeType5Resource, PmodSPIType2AResource,
	PmodSPIType2Resource, PmodUARTType3Resource,
	PmodUARTType4AResource, PmodUARTType4Resource
)
from .interface  import (
	CANResource, DirectUSBResource, I2CResource, IrDAResource,
	JTAGResource, PS2Resource, SPIResource, UARTResource,
	ULPIResource
)
from .memory     import (
	DDR3Resource, NORFlashResources, QSPIFlashResource,
	SDCardResources, SDRAMResource, SPIFlashResources,
	SRAMResource
)
from .user       import (
	ButtonResources, LEDResources, RGBLEDResource, SwitchResources
)

__all__ = (
	'Display7SegResource',
	'VGADACResource',
	'VGAResource',

	'PmodDualHBridgeType6Resource',
	'PmodGPIOType1Resource',
	'PmodHBridgeType5Resource',
	'PmodSPIType2AResource',
	'PmodSPIType2Resource',
	'PmodUARTType3Resource',
	'PmodUARTType4AResource',
	'PmodUARTType4Resource',

	'CANResource',
	'DirectUSBResource',
	'I2CResource',
	'IrDAResource',
	'JTAGResource',
	'PS2Resource',
	'SPIResource',
	'UARTResource',
	'ULPIResource',

	'DDR3Resource',
	'NORFlashResources',
	'QSPIFlashResource',
	'SDCardResources',
	'SDRAMResource',
	'SPIFlashResources',
	'SRAMResource',

	'ButtonResources',
	'LEDResources',
	'RGBLEDResource',
	'SwitchResources',
)



# SPDX-License-Identifier: BSD-2-Clause

from typing   import Literal, Optional, Union

from ...build import Attrs, Pins, Resource, Subsignal

__all__ = (
	'ButtonResources',
	'LEDResources',
	'RGBLEDResource',
	'SwitchResources',
)

def _SplitResources(
	*args,
	pins: Union[str, list[str], dict[str, str]], invert: bool = False,
	conn: Optional[Union[tuple[str, int], int]] = None,
	attrs: Optional[Attrs] = None, default_name: str, dir: Literal['i', 'o', 'io']
) -> list[Resource]:

	if not isinstance(pins, (str, list, dict)):
		raise TypeError(f'pins is expected to be a \'str\', \'list\', or \'dict\' not {pins!r}')

	if isinstance(pins, str):
		pins = pins.split()

	if isinstance(pins, list):
		pins = dict(enumerate(pins))

	resources = []

	for number, pin in pins.items():
		ios = [Pins(pin, dir = dir, invert = invert, conn = conn)]
		if attrs is not None:
			ios.append(attrs)
		resources.append(Resource.family(*args, number, default_name = default_name, ios = ios))

	return resources


def LEDResources(*args, **kwargs) -> list[Resource]:
	return _SplitResources(*args, **kwargs, default_name = 'led', dir = 'o')


def RGBLEDResource(
	*args,
	r: str, g: str, b: str, invert: bool = False,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> Resource:

	ios = []

	ios.append(Subsignal('r', Pins(r, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('g', Pins(g, dir = 'o', invert = invert, conn = conn, assert_width = 1)))
	ios.append(Subsignal('b', Pins(b, dir = 'o', invert = invert, conn = conn, assert_width = 1)))

	if attrs is not None:
		ios.append(attrs)

	return Resource.family(*args, default_name = 'rgb_led', ios = ios)


def ButtonResources(*args, **kwargs) -> list[Resource]:
	return _SplitResources(*args, **kwargs, default_name = 'button', dir = 'i')


def SwitchResources(*args, **kwargs) -> list[Resource]:
	return _SplitResources(*args, **kwargs, default_name = 'switch', dir = 'i')



# SPDX-License-Identifier: BSD-2-Clause

from torii.util import units as util_units
from torii.test import ToriiTestCase

class UnitUtilTestCase(ToriiTestCase):

	def test_time_conversion(self):
		self.assertEqual(util_units.ns_to_sec(183.0), 0.000000183)
		self.assertEqual(util_units.sec_to_ns(0.000000183), 183)
		self.assertEqual(util_units.us_to_sec(24), 0.000024)
		self.assertEqual(util_units.sec_to_us(0.000024), 24)
		self.assertEqual(util_units.ms_to_sec(32.5), 0.0325)
		self.assertEqual(util_units.sec_to_ms(0.0325), 32.5)

	def test_iec_size(self):
		self.assertEqual(util_units.iec_size(1032), '1.01KiB')

		self.assertEqual(util_units.iec_size(256, dec = 0), '256B')

		self.assertEqual(
			util_units.iec_size(18446744073509553615, dec = 10),
			'15.9999999998EiB'
		)

	def test_log2_int(self):
		self.assertEqual(util_units.log2_int(0), 0)
		self.assertEqual(util_units.log2_int(4), 2)
		self.assertEqual(util_units.log2_int(9, False), 4)

		with self.assertRaisesRegex(ValueError, r'^\d is not a power of 2$'):
			util_units.log2_int(3)

	def test_bits_for(self):
		self.assertEqual(util_units.bits_for(1024), 11)
		self.assertEqual(util_units.bits_for(-3), 3)



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, ClockDomain
from torii.back import verilog

class ClockDivisor(Elaboratable):
	def __init__(self, factor: int):
		self.v = Signal(factor)
		self.o = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		m.d.sync += self.v.eq(self.v + 1)
		m.d.comb += self.o.eq(self.v[-1])
		return m


if __name__ == '__main__':
	m = Module()
	m.domains.sync = sync = ClockDomain('sync', async_reset = True)
	m.submodules.ctr = ctr = ClockDivisor(factor = 16)

	print(verilog.convert(m, ports = [ctr.o, sync.clk]))



# SPDX-License-Identifier: BSD-3-Clause


from collections import OrderedDict
from functools   import wraps
from warnings    import warn

__all__ = (
	'deprecated',
	'extend',
	'final',
	'memoize',
)

def memoize(f):
	memo = OrderedDict()

	@wraps(f)
	def g(*args):
		if args not in memo:
			memo[args] = f(*args)
		return memo[args]
	return g

def final(cls):
	def init_subclass():
		raise TypeError(f'Subclassing {cls.__module__}.{cls.__name__} is not supported')
	cls.__init_subclass__ = init_subclass
	return cls

def deprecated(message: str, stacklevel: int = 2):
	def decorator(f):
		@wraps(f)
		def wrapper(*args, **kwargs):
			warn(message, DeprecationWarning, stacklevel = stacklevel)
			return f(*args, **kwargs)
		return wrapper
	return decorator

def extend(cls):
	def decorator(f):
		if isinstance(f, property):
			name = f.fget.__name__
		else:
			name = f.__name__
		setattr(cls, name, f)
	return decorator
